@@ -1,9 +1,7 @@
-from __future__ import absolute_import, division, print_function
-
 from . import config, datasets
 from .core import istask
-from .context import set_options
 from .local import get_sync as get
+
 try:
     from .delayed import delayed
 except ImportError:
@@ -14,7 +12,8 @@ except ImportError:
     pass
 
 from ._version import get_versions
+
 versions = get_versions()
-__version__ = versions['version']
-__git_revision__ = versions['full-revisionid']
+__version__ = versions["version"]
+__git_revision__ = versions["full-revisionid"]
 del get_versions, versions
