@@ -1,21 +1,29 @@
-from __future__ import absolute_import, division, print_function
-
-from functools import partial, wraps
+from collections.abc import Sequence
+from functools import partial, reduce
 from itertools import product
-from operator import add
+from operator import add, getitem
 from numbers import Integral, Number
 
 import numpy as np
 from toolz import accumulate, sliding_window
 
-from .. import sharedict
+from ..highlevelgraph import HighLevelGraph
 from ..base import tokenize
-from ..compatibility import Sequence
+from ..utils import derived_from
 from . import chunk
-from .core import (Array, asarray, normalize_chunks,
-                   stack, concatenate, block,
-                   broadcast_to, broadcast_arrays)
+from .core import (
+    Array,
+    asarray,
+    normalize_chunks,
+    stack,
+    concatenate,
+    block,
+    broadcast_to,
+    broadcast_arrays,
+    cached_cumsum,
+)
 from .wrap import empty, ones, zeros, full
+from .utils import AxisError, meta_from_array, zeros_like_safe
 
 
 def empty_like(a, dtype=None, chunks=None):
@@ -56,8 +64,9 @@ def empty_like(a, dtype=None, chunks=None):
 
     a = asarray(a, name=False)
     return empty(
-        a.shape, dtype=(dtype or a.dtype),
-        chunks=(chunks if chunks is not None else a.chunks)
+        a.shape,
+        dtype=(dtype or a.dtype),
+        chunks=(chunks if chunks is not None else a.chunks),
     )
 
 
@@ -92,8 +101,9 @@ def ones_like(a, dtype=None, chunks=None):
 
     a = asarray(a, name=False)
     return ones(
-        a.shape, dtype=(dtype or a.dtype),
-        chunks=(chunks if chunks is not None else a.chunks)
+        a.shape,
+        dtype=(dtype or a.dtype),
+        chunks=(chunks if chunks is not None else a.chunks),
     )
 
 
@@ -128,8 +138,9 @@ def zeros_like(a, dtype=None, chunks=None):
 
     a = asarray(a, name=False)
     return zeros(
-        a.shape, dtype=(dtype or a.dtype),
-        chunks=(chunks if chunks is not None else a.chunks)
+        a.shape,
+        dtype=(dtype or a.dtype),
+        chunks=(chunks if chunks is not None else a.chunks),
     )
 
 
@@ -171,12 +182,13 @@ def full_like(a, fill_value, dtype=None, chunks=None):
         a.shape,
         fill_value,
         dtype=(dtype or a.dtype),
-        chunks=(chunks if chunks is not None else a.chunks)
+        chunks=(chunks if chunks is not None else a.chunks),
     )
 
 
-def linspace(start, stop, num=50, endpoint=True, retstep=False, chunks=None,
-             dtype=None):
+def linspace(
+    start, stop, num=50, endpoint=True, retstep=False, chunks="auto", dtype=None
+):
     """
     Return `num` evenly spaced values over the closed interval [`start`,
     `stop`].
@@ -200,7 +212,7 @@ def linspace(start, stop, num=50, endpoint=True, retstep=False, chunks=None,
         The number of samples on each block. Note that the last block will have
         fewer samples if `num % blocksize != 0`
     dtype : dtype, optional
-        The type of the output array. Default is given by ``numpy.dtype(float)``.
+        The type of the output array.
 
     Returns
     -------
@@ -215,20 +227,17 @@ def linspace(start, stop, num=50, endpoint=True, retstep=False, chunks=None,
     """
     num = int(num)
 
-    if chunks is None:
-        raise ValueError("Must supply a chunks= keyword argument")
+    if dtype is None:
+        dtype = np.linspace(0, 1, 1).dtype
 
-    chunks = normalize_chunks(chunks, (num,))
+    chunks = normalize_chunks(chunks, (num,), dtype=dtype)
 
     range_ = stop - start
 
     div = (num - 1) if endpoint else num
     step = float(range_) / div
 
-    if dtype is None:
-        dtype = np.linspace(0, 1, 1).dtype
-
-    name = 'linspace-' + tokenize((start, stop, num, endpoint, chunks, dtype))
+    name = "linspace-" + tokenize((start, stop, num, endpoint, chunks, dtype))
 
     dsk = {}
     blockstart = start
@@ -236,8 +245,12 @@ def linspace(start, stop, num=50, endpoint=True, retstep=False, chunks=None,
     for i, bs in enumerate(chunks[0]):
         bs_space = bs - 1 if endpoint else bs
         blockstop = blockstart + (bs_space * step)
-        task = (partial(np.linspace, endpoint=endpoint, dtype=dtype),
-                blockstart, blockstop, bs)
+        task = (
+            partial(np.linspace, endpoint=endpoint, dtype=dtype),
+            blockstart,
+            blockstop,
+            bs,
+        )
         blockstart = blockstart + (step * bs)
         dsk[(name, i)] = task
 
@@ -292,26 +305,26 @@ def arange(*args, **kwargs):
     elif len(args) == 3:
         start, stop, step = args
     else:
-        raise TypeError('''
+        raise TypeError(
+            """
         arange takes 3 positional arguments: arange([start], stop, [step])
-        ''')
+        """
+        )
 
-    try:
-        chunks = kwargs.pop('chunks')
-    except KeyError:
-        raise TypeError("Required argument 'chunks' not found")
+    chunks = kwargs.pop("chunks", "auto")
 
     num = int(max(np.ceil((stop - start) / step), 0))
-    chunks = normalize_chunks(chunks, (num,))
 
-    dtype = kwargs.pop('dtype', None)
+    dtype = kwargs.pop("dtype", None)
     if dtype is None:
         dtype = np.arange(start, stop, step * num if num else step).dtype
-    if kwargs:
-        raise TypeError("Unexpected keyword argument(s): %s" %
-                        ",".join(kwargs.keys()))
 
-    name = 'arange-' + tokenize((start, stop, step, chunks, dtype))
+    chunks = normalize_chunks(chunks, (num,), dtype=dtype)
+
+    if kwargs:
+        raise TypeError("Unexpected keyword argument(s): %s" % ",".join(kwargs.keys()))
+
+    name = "arange-" + tokenize((start, stop, step, chunks, dtype))
     dsk = {}
     elem_count = 0
 
@@ -325,7 +338,7 @@ def arange(*args, **kwargs):
     return Array(dsk, name, chunks, dtype=dtype)
 
 
-@wraps(np.meshgrid)
+@derived_from(np)
 def meshgrid(*xi, **kwargs):
     indexing = kwargs.pop("indexing", "xy")
     sparse = bool(kwargs.pop("sparse", False))
@@ -364,7 +377,7 @@ def meshgrid(*xi, **kwargs):
     return grid
 
 
-def indices(dimensions, dtype=int, chunks=None):
+def indices(dimensions, dtype=int, chunks="auto"):
     """
     Implements NumPy's ``indices`` for Dask Arrays.
 
@@ -380,20 +393,23 @@ def indices(dimensions, dtype=int, chunks=None):
         The shape of the index grid.
     dtype : dtype, optional
         Type to use for the array. Default is ``int``.
-    chunks : sequence of ints
-        The number of samples on each block. Note that the last block will have
-        fewer samples if ``len(array) % chunks != 0``.
+    chunks : sequence of ints, str
+        The size of each block.  Must be one of the following forms:
+
+        -   A blocksize like (500, 1000)
+        -   A size in bytes, like "100 MiB" which will choose a uniform
+            block-like shape
+        -   The word "auto" which acts like the above, but uses a configuration
+            value ``array.chunk-size`` for the chunk size
+        Note that the last block will have fewer samples if ``len(array) % chunks != 0``.
 
     Returns
     -------
     grid : dask array
     """
-    if chunks is None:
-        raise ValueError("Must supply a chunks= keyword argument")
-
     dimensions = tuple(dimensions)
     dtype = np.dtype(dtype)
-    chunks = tuple(chunks)
+    chunks = normalize_chunks(chunks, shape=dimensions, dtype=dtype)
 
     if len(dimensions) != len(chunks):
         raise ValueError("Need same number of chunks as dimensions.")
@@ -409,14 +425,12 @@ def indices(dimensions, dtype=int, chunks=None):
     if grid:
         grid = stack(grid)
     else:
-        grid = empty(
-            (len(dimensions),) + dimensions, dtype=dtype, chunks=(1,) + chunks
-        )
+        grid = empty((len(dimensions),) + dimensions, dtype=dtype, chunks=(1,) + chunks)
 
     return grid
 
 
-def eye(N, chunks, M=None, k=0, dtype=float):
+def eye(N, chunks="auto", M=None, k=0, dtype=float):
     """
     Return a 2-D Array with ones on the diagonal and zeros elsewhere.
 
@@ -424,8 +438,14 @@ def eye(N, chunks, M=None, k=0, dtype=float):
     ----------
     N : int
       Number of rows in the output.
-    chunks: int
-        chunk size of resulting blocks
+    chunks : int, str
+        How to chunk the array. Must be one of the following forms:
+
+        -   A blocksize like 1000.
+        -   A size in bytes, like "100 MiB" which will choose a uniform
+            block-like shape
+        -   The word "auto" which acts like the above, but uses a configuration
+            value ``array.chunk-size`` for the chunk size
     M : int, optional
       Number of columns in the output. If None, defaults to `N`.
     k : int, optional
@@ -441,16 +461,18 @@ def eye(N, chunks, M=None, k=0, dtype=float):
       An array where all elements are equal to zero, except for the `k`-th
       diagonal, whose values are equal to one.
     """
-    if not isinstance(chunks, Integral):
-        raise ValueError('chunks must be an int')
-
-    token = tokenize(N, chunk, M, k, dtype)
-    name_eye = 'eye-' + token
-
     eye = {}
     if M is None:
         M = N
 
+    if not isinstance(chunks, (int, str)):
+        raise ValueError("chunks must be an int or string")
+    elif isinstance(chunks, str):
+        chunks = normalize_chunks(chunks, shape=(N, M), dtype=dtype)
+        chunks = chunks[0][0]
+    token = tokenize(N, chunks, M, k, dtype)
+    name_eye = "eye-" + token
+
     vchunks = [chunks] * (N // chunks)
     if N % chunks != 0:
         vchunks.append(N % chunks)
@@ -461,17 +483,27 @@ def eye(N, chunks, M=None, k=0, dtype=float):
     for i, vchunk in enumerate(vchunks):
         for j, hchunk in enumerate(hchunks):
             if (j - i - 1) * chunks <= k <= (j - i + 1) * chunks:
-                eye[name_eye, i, j] = (np.eye, vchunk, hchunk, k - (j - i) * chunks, dtype)
+                eye[name_eye, i, j] = (
+                    np.eye,
+                    vchunk,
+                    hchunk,
+                    k - (j - i) * chunks,
+                    dtype,
+                )
             else:
                 eye[name_eye, i, j] = (np.zeros, (vchunk, hchunk), dtype)
-    return Array(eye, name_eye, shape=(N, M),
-                 chunks=(chunks, chunks), dtype=dtype)
+    return Array(eye, name_eye, shape=(N, M), chunks=(chunks, chunks), dtype=dtype)
 
 
-@wraps(np.diag)
+@derived_from(np)
 def diag(v):
-    name = 'diag-' + tokenize(v)
-    if isinstance(v, np.ndarray):
+    name = "diag-" + tokenize(v)
+
+    meta = meta_from_array(v, 2 if v.ndim == 1 else 1)
+
+    if isinstance(v, np.ndarray) or (
+        hasattr(v, "__array_function__") and not isinstance(v, Array)
+    ):
         if v.ndim == 1:
             chunks = ((v.shape[0],), (v.shape[0],))
             dsk = {(name, 0, 0): (np.diag, v)}
@@ -480,20 +512,22 @@ def diag(v):
             dsk = {(name, 0): (np.diag, v)}
         else:
             raise ValueError("Array must be 1d or 2d only")
-        return Array(dsk, name, chunks, dtype=v.dtype)
+        return Array(dsk, name, chunks, meta=meta)
     if not isinstance(v, Array):
-        raise TypeError("v must be a dask array or numpy array, "
-                        "got {0}".format(type(v)))
+        raise TypeError(
+            "v must be a dask array or numpy array, got {0}".format(type(v))
+        )
     if v.ndim != 1:
         if v.chunks[0] == v.chunks[1]:
-            dsk = {(name, i): (np.diag, row[i])
-                   for i, row in enumerate(v.__dask_keys__())}
-            return Array(sharedict.merge(v.dask, (name, dsk),
-                                         dependencies={name: {v.name}}),
-                         name, (v.chunks[0],), dtype=v.dtype)
+            dsk = {
+                (name, i): (np.diag, row[i]) for i, row in enumerate(v.__dask_keys__())
+            }
+            graph = HighLevelGraph.from_collections(name, dsk, dependencies=[v])
+            return Array(graph, name, (v.chunks[0],), meta=meta)
         else:
-            raise NotImplementedError("Extracting diagonals from non-square "
-                                      "chunked arrays")
+            raise NotImplementedError(
+                "Extracting diagonals from non-square chunked arrays"
+            )
     chunks_1d = v.chunks[0]
     blocks = v.__dask_keys__()
     dsk = {}
@@ -504,10 +538,78 @@ def diag(v):
                 dsk[key] = (np.diag, blocks[i])
             else:
                 dsk[key] = (np.zeros, (m, n))
+                dsk[key] = (partial(zeros_like_safe, shape=(m, n)), meta)
 
-    return Array(sharedict.merge(v.dask, (name, dsk),
-                                 dependencies={name: {v.name}}),
-                 name, (chunks_1d, chunks_1d), dtype=v.dtype)
+    graph = HighLevelGraph.from_collections(name, dsk, dependencies=[v])
+    return Array(graph, name, (chunks_1d, chunks_1d), meta=meta)
+
+
+@derived_from(np)
+def diagonal(a, offset=0, axis1=0, axis2=1):
+    name = "diagonal-" + tokenize(a, offset, axis1, axis2)
+
+    if a.ndim < 2:
+        # NumPy uses `diag` as we do here.
+        raise ValueError("diag requires an array of at least two dimensions")
+
+    def _axis_fmt(axis, name, ndim):
+        if axis < 0:
+            t = ndim + axis
+            if t < 0:
+                msg = "{}: axis {} is out of bounds for array of dimension {}"
+                raise AxisError(msg.format(name, axis, ndim))
+            axis = t
+        return axis
+
+    axis1 = _axis_fmt(axis1, "axis1", a.ndim)
+    axis2 = _axis_fmt(axis2, "axis2", a.ndim)
+
+    if axis1 == axis2:
+        raise ValueError("axis1 and axis2 cannot be the same")
+
+    a = asarray(a)
+
+    if axis1 > axis2:
+        axis1, axis2 = axis2, axis1
+        offset = -offset
+
+    def _diag_len(dim1, dim2, offset):
+        return max(0, min(min(dim1, dim2), dim1 + offset, dim2 - offset))
+
+    diag_chunks = []
+    chunk_offsets = []
+    cum1 = list(cached_cumsum(a.chunks[axis1], initial_zero=True)[:-1])
+    cum2 = list(cached_cumsum(a.chunks[axis2], initial_zero=True)[:-1])
+    for co1, c1 in zip(cum1, a.chunks[axis1]):
+        chunk_offsets.append([])
+        for co2, c2 in zip(cum2, a.chunks[axis2]):
+            k = offset + co1 - co2
+            diag_chunks.append(_diag_len(c1, c2, k))
+            chunk_offsets[-1].append(k)
+
+    dsk = {}
+    idx_set = set(range(a.ndim)) - set([axis1, axis2])
+    n1 = len(a.chunks[axis1])
+    n2 = len(a.chunks[axis2])
+    for idx in product(*(range(len(a.chunks[i])) for i in idx_set)):
+        for i, (i1, i2) in enumerate(product(range(n1), range(n2))):
+            tsk = reduce(getitem, idx[:axis1], a.__dask_keys__())[i1]
+            tsk = reduce(getitem, idx[axis1 : axis2 - 1], tsk)[i2]
+            tsk = reduce(getitem, idx[axis2 - 1 :], tsk)
+            k = chunk_offsets[i1][i2]
+            dsk[(name,) + idx + (i,)] = (np.diagonal, tsk, k, axis1, axis2)
+
+    left_shape = tuple(a.shape[i] for i in idx_set)
+    right_shape = (_diag_len(a.shape[axis1], a.shape[axis2], offset),)
+    shape = left_shape + right_shape
+
+    left_chunks = tuple(a.chunks[i] for i in idx_set)
+    right_shape = (tuple(diag_chunks),)
+    chunks = left_chunks + right_shape
+
+    graph = HighLevelGraph.from_collections(name, dsk, dependencies=[a])
+    meta = meta_from_array(a, len(shape))
+    return Array(graph, name, shape=shape, chunks=chunks, meta=meta)
 
 
 def triu(m, k=0):
@@ -532,10 +634,12 @@ def triu(m, k=0):
     tril : lower triangle of an array
     """
     if m.ndim != 2:
-        raise ValueError('input must be 2 dimensional')
+        raise ValueError("input must be 2 dimensional")
     if m.chunks[0][0] != m.chunks[1][0]:
-        msg = ('chunks must be a square. '
-               'Use .rechunk method to change the size of chunks.')
+        msg = (
+            "chunks must be a square. "
+            "Use .rechunk method to change the size of chunks."
+        )
         raise NotImplementedError(msg)
 
     rdim = len(m.chunks[0])
@@ -543,19 +647,22 @@ def triu(m, k=0):
     chunk = m.chunks[0][0]
 
     token = tokenize(m, k)
-    name = 'triu-' + token
+    name = "triu-" + token
 
     dsk = {}
     for i in range(rdim):
         for j in range(hdim):
             if chunk * (j - i + 1) < k:
-                dsk[(name, i, j)] = (np.zeros, (m.chunks[0][i], m.chunks[1][j]))
+                dsk[(name, i, j)] = (
+                    partial(zeros_like_safe, shape=(m.chunks[0][i], m.chunks[1][j])),
+                    m._meta,
+                )
             elif chunk * (j - i - 1) < k <= chunk * (j - i + 1):
                 dsk[(name, i, j)] = (np.triu, (m.name, i, j), k - (chunk * (j - i)))
             else:
                 dsk[(name, i, j)] = (m.name, i, j)
-    return Array(sharedict.merge((name, dsk), m.dask, dependencies={name: {m.name}}),
-                 name, shape=m.shape, chunks=m.chunks, dtype=m.dtype)
+    graph = HighLevelGraph.from_collections(name, dsk, dependencies=[m])
+    return Array(graph, name, shape=m.shape, chunks=m.chunks, meta=m)
 
 
 def tril(m, k=0):
@@ -580,10 +687,12 @@ def tril(m, k=0):
     triu : upper triangle of an array
     """
     if m.ndim != 2:
-        raise ValueError('input must be 2 dimensional')
+        raise ValueError("input must be 2 dimensional")
     if not len(set(m.chunks[0] + m.chunks[1])) == 1:
-        msg = ('All chunks must be a square matrix to perform lu decomposition. '
-               'Use .rechunk method to change the size of chunks.')
+        msg = (
+            "All chunks must be a square matrix to perform lu decomposition. "
+            "Use .rechunk method to change the size of chunks."
+        )
         raise ValueError(msg)
 
     rdim = len(m.chunks[0])
@@ -591,7 +700,7 @@ def tril(m, k=0):
     chunk = m.chunks[0][0]
 
     token = tokenize(m, k)
-    name = 'tril-' + token
+    name = "tril-" + token
 
     dsk = {}
     for i in range(rdim):
@@ -601,9 +710,12 @@ def tril(m, k=0):
             elif chunk * (j - i - 1) < k <= chunk * (j - i + 1):
                 dsk[(name, i, j)] = (np.tril, (m.name, i, j), k - (chunk * (j - i)))
             else:
-                dsk[(name, i, j)] = (np.zeros, (m.chunks[0][i], m.chunks[1][j]))
-    dsk = sharedict.merge(m.dask, (name, dsk), dependencies={name: {m.name}})
-    return Array(dsk, name, shape=m.shape, chunks=m.chunks, dtype=m.dtype)
+                dsk[(name, i, j)] = (
+                    partial(zeros_like_safe, shape=(m.chunks[0][i], m.chunks[1][j])),
+                    m._meta,
+                )
+    graph = HighLevelGraph.from_collections(name, dsk, dependencies=[m])
+    return Array(graph, name, shape=m.shape, chunks=m.chunks, meta=m)
 
 
 def _np_fromfunction(func, shape, dtype, offset, func_kwargs):
@@ -614,26 +726,27 @@ def _np_fromfunction(func, shape, dtype, offset, func_kwargs):
     return np.fromfunction(offset_func, shape, dtype=dtype, **func_kwargs)
 
 
-@wraps(np.fromfunction)
-def fromfunction(func, chunks=None, shape=None, dtype=None, **kwargs):
-    if chunks:
-        chunks = normalize_chunks(chunks, shape)
-    name = 'fromfunction-' + tokenize(func, chunks, shape, dtype, kwargs)
+@derived_from(np)
+def fromfunction(func, chunks="auto", shape=None, dtype=None, **kwargs):
+    chunks = normalize_chunks(chunks, shape, dtype=dtype)
+    name = "fromfunction-" + tokenize(func, chunks, shape, dtype, kwargs)
     keys = list(product([name], *[range(len(bd)) for bd in chunks]))
     aggdims = [list(accumulate(add, (0,) + bd[:-1])) for bd in chunks]
     offsets = list(product(*aggdims))
     shapes = list(product(*chunks))
     dtype = dtype or float
 
-    values = [(_np_fromfunction, func, shp, dtype, offset, kwargs)
-              for offset, shp in zip(offsets, shapes)]
+    values = [
+        (_np_fromfunction, func, shp, dtype, offset, kwargs)
+        for offset, shp in zip(offsets, shapes)
+    ]
 
     dsk = dict(zip(keys, values))
 
     return Array(dsk, name, chunks, dtype=dtype)
 
 
-@wraps(np.repeat)
+@derived_from(np)
 def repeat(a, repeats, axis=None):
     if axis is None:
         if a.ndim == 1:
@@ -652,7 +765,7 @@ def repeat(a, repeats, axis=None):
     if repeats == 1:
         return a
 
-    cchunks = np.cumsum((0,) + a.chunks[axis])
+    cchunks = cached_cumsum(a.chunks[axis], initial_zero=True)
     slices = []
     for c_start, c_stop in sliding_window(2, cchunks):
         ls = np.linspace(c_start, c_stop, repeats).round(0)
@@ -661,8 +774,9 @@ def repeat(a, repeats, axis=None):
                 slices.append(slice(ls_start, ls_stop))
 
     all_slice = slice(None, None, None)
-    slices = [(all_slice,) * axis + (s,) + (all_slice,) * (a.ndim - axis - 1)
-              for s in slices]
+    slices = [
+        (all_slice,) * axis + (s,) + (all_slice,) * (a.ndim - axis - 1) for s in slices
+    ]
 
     slabs = [a[slc] for slc in slices]
 
@@ -672,51 +786,65 @@ def repeat(a, repeats, axis=None):
         assert len(chunks[axis]) == 1
         chunks[axis] = (chunks[axis][0] * repeats,)
         chunks = tuple(chunks)
-        result = slab.map_blocks(np.repeat, repeats, axis=axis, chunks=chunks,
-                                 dtype=slab.dtype)
+        result = slab.map_blocks(
+            np.repeat, repeats, axis=axis, chunks=chunks, dtype=slab.dtype
+        )
         out.append(result)
 
     return concatenate(out, axis=axis)
 
 
-@wraps(np.tile)
+@derived_from(np)
 def tile(A, reps):
-    if not isinstance(reps, Integral):
-        raise NotImplementedError("Only integer valued `reps` supported.")
-
-    if reps < 0:
+    try:
+        tup = tuple(reps)
+    except TypeError:
+        tup = (reps,)
+    if any(i < 0 for i in tup):
         raise ValueError("Negative `reps` are not allowed.")
-    elif reps == 0:
-        return A[..., :0]
-    elif reps == 1:
-        return A
+    c = asarray(A)
 
-    return concatenate(reps * [A], axis=-1)
+    if all(tup):
+        for nrep in tup[::-1]:
+            c = nrep * [c]
+        return block(c)
+
+    d = len(tup)
+    if d < c.ndim:
+        tup = (1,) * (c.ndim - d) + tup
+    if c.ndim < d:
+        shape = (1,) * (d - c.ndim) + c.shape
+    else:
+        shape = c.shape
+    shape_out = tuple(s * t for s, t in zip(shape, tup))
+    return empty(shape=shape_out, dtype=c.dtype)
 
 
 def expand_pad_value(array, pad_value):
     if isinstance(pad_value, Number):
         pad_value = array.ndim * ((pad_value, pad_value),)
-    elif (isinstance(pad_value, Sequence) and
-          all(isinstance(pw, Number) for pw in pad_value) and
-          len(pad_value) == 1):
+    elif (
+        isinstance(pad_value, Sequence)
+        and all(isinstance(pw, Number) for pw in pad_value)
+        and len(pad_value) == 1
+    ):
         pad_value = array.ndim * ((pad_value[0], pad_value[0]),)
-    elif (isinstance(pad_value, Sequence) and
-          len(pad_value) == 2 and
-          all(isinstance(pw, Number) for pw in pad_value)):
-            pad_value = tuple(
-                (pad_value[0], pad_value[1]) for _ in range(array.ndim)
-            )
-    elif (isinstance(pad_value, Sequence) and
-          len(pad_value) == array.ndim and
-          all(isinstance(pw, Sequence) for pw in pad_value) and
-          all((len(pw) == 2) for pw in pad_value) and
-          all(all(isinstance(w, Number) for w in pw) for pw in pad_value)):
-            pad_value = tuple((pw[0], pw[1]) for pw in pad_value)
+    elif (
+        isinstance(pad_value, Sequence)
+        and len(pad_value) == 2
+        and all(isinstance(pw, Number) for pw in pad_value)
+    ):
+        pad_value = tuple((pad_value[0], pad_value[1]) for _ in range(array.ndim))
+    elif (
+        isinstance(pad_value, Sequence)
+        and len(pad_value) == array.ndim
+        and all(isinstance(pw, Sequence) for pw in pad_value)
+        and all((len(pw) == 2) for pw in pad_value)
+        and all(all(isinstance(w, Number) for w in pw) for pw in pad_value)
+    ):
+        pad_value = tuple((pw[0], pw[1]) for pw in pad_value)
     else:
-        raise TypeError(
-            "`pad_value` must be composed of integral typed values."
-        )
+        raise TypeError("`pad_value` must be composed of integral typed values.")
 
     return pad_value
 
@@ -778,38 +906,39 @@ def pad_edge(array, pad_width, mode, *args):
         pad_shapes, pad_chunks = get_pad_shapes_chunks(result, pad_width, (d,))
         pad_arrays = [result, result]
 
-        if mode is "constant":
+        if mode == "constant":
             constant_values = args[0][d]
-            constant_values = [
-                asarray(c).astype(result.dtype) for c in constant_values
-            ]
+            constant_values = [asarray(c).astype(result.dtype) for c in constant_values]
 
             pad_arrays = [
                 broadcast_to(v, s, c)
                 for v, s, c in zip(constant_values, pad_shapes, pad_chunks)
             ]
         elif mode in ["edge", "linear_ramp"]:
-            pad_slices = [
-                result.ndim * [slice(None)], result.ndim * [slice(None)]
-            ]
+            pad_slices = [result.ndim * [slice(None)], result.ndim * [slice(None)]]
             pad_slices[0][d] = slice(None, 1, None)
             pad_slices[1][d] = slice(-1, None, None)
             pad_slices = [tuple(sl) for sl in pad_slices]
 
             pad_arrays = [result[sl] for sl in pad_slices]
 
-            if mode is "edge":
+            if mode == "edge":
                 pad_arrays = [
                     broadcast_to(a, s, c)
                     for a, s, c in zip(pad_arrays, pad_shapes, pad_chunks)
                 ]
-            elif mode is "linear_ramp":
+            elif mode == "linear_ramp":
                 end_values = args[0][d]
 
                 pad_arrays = [
                     a.map_blocks(
-                        linear_ramp_chunk, ev, pw,
-                        chunks=c, dtype=result.dtype, dim=d, step=(2 * i - 1)
+                        linear_ramp_chunk,
+                        ev,
+                        pw,
+                        chunks=c,
+                        dtype=result.dtype,
+                        dim=d,
+                        step=(2 * i - 1),
                     )
                     for i, (a, ev, pw, c) in enumerate(
                         zip(pad_arrays, end_values, pad_width[d], pad_chunks)
@@ -831,9 +960,7 @@ def pad_reuse(array, pad_width, mode, *args):
     """
 
     if mode in ["reflect", "symmetric"] and "odd" in args:
-        raise NotImplementedError(
-            "`pad` does not support `reflect_type` of `odd`."
-        )
+        raise NotImplementedError("`pad` does not support `reflect_type` of `odd`.")
 
     result = np.empty(array.ndim * (3,), dtype=object)
     for idx in np.ndindex(result.shape):
@@ -894,9 +1021,9 @@ def pad_stats(array, pad_width, mode, *args):
         select = []
         pad_shape = []
         pad_chunks = []
-        for d, (i, s, c, w, l) in enumerate(zip(
-            idx, array.shape, array.chunks, pad_width, stat_length
-        )):
+        for d, (i, s, c, w, l) in enumerate(
+            zip(idx, array.shape, array.chunks, pad_width, stat_length)
+        ):
             if i < 1:
                 axes.append(d)
                 select.append(slice(None, l[0], None))
@@ -918,14 +1045,15 @@ def pad_stats(array, pad_width, mode, *args):
         pad_chunks = tuple(pad_chunks)
 
         result_idx = array[select]
-        if mode == "maximum":
-            result_idx = result_idx.max(axis=axes, keepdims=True)
-        elif mode == "mean":
-            result_idx = result_idx.mean(axis=axes, keepdims=True)
-        elif mode == "minimum":
-            result_idx = result_idx.min(axis=axes, keepdims=True)
+        if axes:
+            if mode == "maximum":
+                result_idx = result_idx.max(axis=axes, keepdims=True)
+            elif mode == "mean":
+                result_idx = result_idx.mean(axis=axes, keepdims=True)
+            elif mode == "minimum":
+                result_idx = result_idx.min(axis=axes, keepdims=True)
 
-        result_idx = broadcast_to(result_idx, pad_shape, chunks=pad_chunks)
+            result_idx = broadcast_to(result_idx, pad_shape, chunks=pad_chunks)
 
         result[idx] = result_idx
 
@@ -936,7 +1064,7 @@ def pad_stats(array, pad_width, mode, *args):
 
 def wrapped_pad_func(array, pad_func, iaxis_pad_width, iaxis, pad_func_kwargs):
     result = np.empty_like(array)
-    for i in np.ndindex(array.shape[:iaxis] + array.shape[iaxis + 1:]):
+    for i in np.ndindex(array.shape[:iaxis] + array.shape[iaxis + 1 :]):
         i = i[:iaxis] + (slice(None),) + i[iaxis:]
         result[i] = pad_func(array[i], iaxis_pad_width, iaxis, pad_func_kwargs)
 
@@ -958,7 +1086,7 @@ def pad_udf(array, pad_width, mode, **kwargs):
     chunks = result.chunks
     for d in range(result.ndim):
         result = result.rechunk(
-            chunks[:d] + (result.shape[d:d + 1],) + chunks[d + 1:]
+            chunks[:d] + (result.shape[d : d + 1],) + chunks[d + 1 :]
         )
 
         result = result.map_blocks(
@@ -976,7 +1104,7 @@ def pad_udf(array, pad_width, mode, **kwargs):
     return result
 
 
-@wraps(np.pad)
+@derived_from(np)
 def pad(array, pad_width, mode, **kwargs):
     array = asarray(array)
 
