@@ -1,15 +1,12 @@
-from __future__ import absolute_import, division, print_function
-
 from operator import getitem
 from itertools import product
 from numbers import Integral
-
 from toolz import merge, pipe, concat, partial
 from toolz.curried import map
 
 from . import chunk, wrap
 from .core import Array, map_blocks, concatenate, concatenate3, reshapelist
-from .. import sharedict
+from ..highlevelgraph import HighLevelGraph
 from ..base import tokenize
 from ..core import flatten
 from ..utils import concrete
@@ -32,15 +29,21 @@ def fractional_slice(task, axes):
     index = []
     for i, (t, r) in enumerate(zip(task[1:], rounded[1:])):
         depth = axes.get(i, 0)
+        if isinstance(depth, tuple):
+            left_depth = depth[0]
+            right_depth = depth[1]
+        else:
+            left_depth = depth
+            right_depth = depth
+
         if t == r:
             index.append(slice(None, None, None))
-        elif t < r:
-            index.append(slice(0, depth))
-        elif t > r and depth == 0:
-            index.append(slice(0, 0))
+        elif t < r and right_depth:
+            index.append(slice(0, right_depth))
+        elif t > r and left_depth:
+            index.append(slice(-left_depth, None))
         else:
-            index.append(slice(-depth, None))
-
+            index.append(slice(0, 0))
     index = tuple(index)
 
     if all(ind == slice(None, None, None) for ind in index):
@@ -74,6 +77,7 @@ def expand_key(k, dims, name=None, axes=None):
     [[('y',   0, 3.1), ('y',   0,   4)],
      [('y', 0.9, 3.1), ('y', 0.9,   4)]]
     """
+
     def inds(i, ind):
         rv = []
         if ind - 0.9 > 0:
@@ -92,11 +96,13 @@ def expand_key(k, dims, name=None, axes=None):
             num += 1
         shape.append(num)
 
-    args = [inds(i, ind) if axes.get(i, 0) else [ind] for i, ind in enumerate(k[1:])]
+    args = [
+        inds(i, ind) if any((axes.get(i, 0),)) else [ind] for i, ind in enumerate(k[1:])
+    ]
     if name is not None:
         args = [[name]] + args
     seq = list(product(*args))
-    shape2 = [d if axes.get(i, 0) else 1 for i, d in enumerate(shape)]
+    shape2 = [d if any((axes.get(i, 0),)) else 1 for i, d in enumerate(shape)]
     result = reshapelist(shape2, seq)
     return result
 
@@ -119,11 +125,12 @@ def overlap_internal(x, axes):
     expand_key2 = partial(expand_key, dims=dims, axes=axes)
 
     # Make keys for each of the surrounding sub-arrays
-    interior_keys = pipe(x.__dask_keys__(), flatten, map(expand_key2),
-                         map(flatten), concat, list)
+    interior_keys = pipe(
+        x.__dask_keys__(), flatten, map(expand_key2), map(flatten), concat, list
+    )
 
-    name = 'overlap-' + tokenize(x, axes)
-    getitem_name = 'getitem-' + tokenize(x, axes)
+    name = "overlap-" + tokenize(x, axes)
+    getitem_name = "getitem-" + tokenize(x, axes)
     interior_slices = {}
     overlap_blocks = {}
     for k in interior_keys:
@@ -132,25 +139,35 @@ def overlap_internal(x, axes):
             interior_slices[(getitem_name,) + k] = frac_slice
         else:
             interior_slices[(getitem_name,) + k] = (x.name,) + k
-            overlap_blocks[(name,) + k] = (concatenate3,
-                                           (concrete, expand_key2((None,) + k, name=getitem_name)))
+            overlap_blocks[(name,) + k] = (
+                concatenate3,
+                (concrete, expand_key2((None,) + k, name=getitem_name)),
+            )
 
     chunks = []
     for i, bds in enumerate(x.chunks):
+        depth = axes.get(i, 0)
+        if isinstance(depth, tuple):
+            left_depth = depth[0]
+            right_depth = depth[1]
+        else:
+            left_depth = depth
+            right_depth = depth
+
         if len(bds) == 1:
             chunks.append(bds)
         else:
-            left = [bds[0] + axes.get(i, 0)]
-            right = [bds[-1] + axes.get(i, 0)]
+            left = [bds[0] + right_depth]
+            right = [bds[-1] + left_depth]
             mid = []
             for bd in bds[1:-1]:
-                mid.append(bd + axes.get(i, 0) * 2)
+                mid.append(bd + left_depth + right_depth)
             chunks.append(left + mid + right)
 
     dsk = merge(interior_slices, overlap_blocks)
-    dsk = sharedict.merge(x.dask, (name, dsk), dependencies={name: {x.name}})
+    graph = HighLevelGraph.from_collections(name, dsk, dependencies=[x])
 
-    return Array(dsk, name, chunks, dtype=x.dtype)
+    return Array(graph, name, chunks, meta=x)
 
 
 def trim_overlap(x, depth, boundary=None):
@@ -186,21 +203,31 @@ def trim_internal(x, axes, boundary=None):
 
     olist = []
     for i, bd in enumerate(x.chunks):
-        bdy = boundary.get(i, 'none')
+        bdy = boundary.get(i, "none")
+        overlap = axes.get(i, 0)
         ilist = []
         for j, d in enumerate(bd):
-            if bdy != 'none':
-                d = d - axes.get(i, 0) * 2
+            if bdy != "none":
+                if isinstance(overlap, tuple):
+                    d = d - sum(overlap)
+                else:
+                    d = d - overlap * 2
+
             else:
-                d = d - axes.get(i, 0) if j != 0 else d
-                d = d - axes.get(i, 0) if j != len(bd) - 1 else d
+                if isinstance(overlap, tuple):
+                    d = d - overlap[0] if j != 0 else d
+                    d = d - overlap[1] if j != len(bd) - 1 else d
+                else:
+                    d = d - overlap if j != 0 else d
+                    d = d - overlap if j != len(bd) - 1 else d
+
             ilist.append(d)
         olist.append(tuple(ilist))
-
     chunks = tuple(olist)
 
-    return map_blocks(partial(_trim, axes=axes, boundary=boundary),
-                      x, chunks=chunks, dtype=x.dtype)
+    return map_blocks(
+        partial(_trim, axes=axes, boundary=boundary), x, chunks=chunks, dtype=x.dtype
+    )
 
 
 def _trim(x, axes, boundary, block_info):
@@ -211,23 +238,32 @@ def _trim(x, axes, boundary, block_info):
 
     """
     axes = [axes.get(i, 0) for i in range(x.ndim)]
-    axes_back = (-ax if ax else None for ax in axes)
+    axes_front = (ax[0] if isinstance(ax, tuple) else ax for ax in axes)
+    axes_back = (
+        -ax[1]
+        if isinstance(ax, tuple) and ax[1]
+        else -ax
+        if isinstance(ax, Integral) and ax
+        else None
+        for ax in axes
+    )
 
     trim_front = (
-        0 if (chunk_location == 0 and
-              boundary.get(i, 'none') == 'none') else ax
+        0 if (chunk_location == 0 and boundary.get(i, "none") == "none") else ax
         for i, (chunk_location, ax) in enumerate(
-            zip(block_info[0]['chunk-location'], axes)))
+            zip(block_info[0]["chunk-location"], axes_front)
+        )
+    )
     trim_back = (
-        None if (chunk_location == chunks - 1 and
-                 boundary.get(i, 'none') == 'none') else ax
-        for i, (chunks, chunk_location, ax) in enumerate(zip(
-            block_info[0]['num-chunks'],
-            block_info[0]['chunk-location'],
-            axes_back)))
-
-    return x[tuple(slice(front, back)
-             for front, back in zip(trim_front, trim_back))]
+        None
+        if (chunk_location == chunks - 1 and boundary.get(i, "none") == "none")
+        else ax
+        for i, (chunks, chunk_location, ax) in enumerate(
+            zip(block_info[0]["num-chunks"], block_info[0]["chunk-location"], axes_back)
+        )
+    )
+    ind = tuple(slice(front, back) for front, back in zip(trim_front, trim_back))
+    return x[ind]
 
 
 def periodic(x, axis, depth):
@@ -236,12 +272,16 @@ def periodic(x, axis, depth):
     Useful to create periodic boundary conditions for overlap
     """
 
-    left = ((slice(None, None, None),) * axis +
-            (slice(0, depth),) +
-            (slice(None, None, None),) * (x.ndim - axis - 1))
-    right = ((slice(None, None, None),) * axis +
-             (slice(-depth, None),) +
-             (slice(None, None, None),) * (x.ndim - axis - 1))
+    left = (
+        (slice(None, None, None),) * axis
+        + (slice(0, depth),)
+        + (slice(None, None, None),) * (x.ndim - axis - 1)
+    )
+    right = (
+        (slice(None, None, None),) * axis
+        + (slice(-depth, None),)
+        + (slice(None, None, None),) * (x.ndim - axis - 1)
+    )
     l = x[left]
     r = x[right]
 
@@ -256,16 +296,22 @@ def reflect(x, axis, depth):
     This is the converse of ``periodic``
     """
     if depth == 1:
-        left = ((slice(None, None, None),) * axis +
-                (slice(0, 1),) +
-                (slice(None, None, None),) * (x.ndim - axis - 1))
+        left = (
+            (slice(None, None, None),) * axis
+            + (slice(0, 1),)
+            + (slice(None, None, None),) * (x.ndim - axis - 1)
+        )
     else:
-        left = ((slice(None, None, None),) * axis +
-                (slice(depth - 1, None, -1),) +
-                (slice(None, None, None),) * (x.ndim - axis - 1))
-    right = ((slice(None, None, None),) * axis +
-             (slice(-1, -depth - 1, -1),) +
-             (slice(None, None, None),) * (x.ndim - axis - 1))
+        left = (
+            (slice(None, None, None),) * axis
+            + (slice(depth - 1, None, -1),)
+            + (slice(None, None, None),) * (x.ndim - axis - 1)
+        )
+    right = (
+        (slice(None, None, None),) * axis
+        + (slice(-1, -depth - 1, -1),)
+        + (slice(None, None, None),) * (x.ndim - axis - 1)
+    )
     l = x[left]
     r = x[right]
 
@@ -280,12 +326,16 @@ def nearest(x, axis, depth):
     This mimics what the skimage.filters.gaussian_filter(... mode="nearest")
     does.
     """
-    left = ((slice(None, None, None),) * axis +
-            (slice(0, 1),) +
-            (slice(None, None, None),) * (x.ndim - axis - 1))
-    right = ((slice(None, None, None),) * axis +
-             (slice(-1, -2, -1),) +
-             (slice(None, None, None),) * (x.ndim - axis - 1))
+    left = (
+        (slice(None, None, None),) * axis
+        + (slice(0, 1),)
+        + (slice(None, None, None),) * (x.ndim - axis - 1)
+    )
+    right = (
+        (slice(None, None, None),) * axis
+        + (slice(-1, -2, -1),)
+        + (slice(None, None, None),) * (x.ndim - axis - 1)
+    )
 
     l = concatenate([x[left]] * depth, axis=axis)
     r = concatenate([x[right]] * depth, axis=axis)
@@ -300,8 +350,18 @@ def constant(x, axis, depth, value):
     chunks = list(x.chunks)
     chunks[axis] = (depth,)
 
-    c = wrap.full(tuple(map(sum, chunks)), value,
-                  chunks=tuple(chunks), dtype=x.dtype)
+    try:
+        c = wrap.full_like(
+            getattr(x, "_meta", x),
+            value,
+            shape=tuple(map(sum, chunks)),
+            chunks=tuple(chunks),
+            dtype=x.dtype,
+        )
+    except TypeError:
+        c = wrap.full(
+            tuple(map(sum, chunks)), value, chunks=tuple(chunks), dtype=x.dtype
+        )
 
     return concatenate([c, x, c], axis=axis)
 
@@ -335,14 +395,14 @@ def boundaries(x, depth=None, kind=None):
         if d == 0:
             continue
 
-        this_kind = kind.get(i, 'none')
-        if this_kind == 'none':
+        this_kind = kind.get(i, "none")
+        if this_kind == "none":
             continue
-        elif this_kind == 'periodic':
+        elif this_kind == "periodic":
             x = periodic(x, i, d)
-        elif this_kind == 'reflect':
+        elif this_kind == "reflect":
             x = reflect(x, i, d)
-        elif this_kind == 'nearest':
+        elif this_kind == "nearest":
             x = nearest(x, i, d)
         elif i in kind:
             x = constant(x, i, d, kind[i])
@@ -408,16 +468,20 @@ def overlap(x, depth, boundary):
     # is depth larger than chunk size?
     depth_values = [depth2.get(i, 0) for i in range(x.ndim)]
     for d, c in zip(depth_values, x.chunks):
-        if d > min(c):
-            raise ValueError("The overlapping depth %d is larger than your\n"
-                             "smallest chunk size %d. Rechunk your array\n"
-                             "with a larger chunk size or a chunk size that\n"
-                             "more evenly divides the shape of your array." %
-                             (d, min(c)))
+        maxd = max(d) if isinstance(d, tuple) else d
+        if maxd > min(c):
+            raise ValueError(
+                "The overlapping depth %d is larger than your\n"
+                "smallest chunk size %d. Rechunk your array\n"
+                "with a larger chunk size or a chunk size that\n"
+                "more evenly divides the shape of your array." % (d, min(c))
+            )
     x2 = boundaries(x, depth2, boundary2)
     x3 = overlap_internal(x2, depth2)
-    trim = dict((k, v * 2 if boundary2.get(k, 'none') != 'none' else 0)
-                for k, v in depth2.items())
+    trim = dict(
+        (k, v * 2 if boundary2.get(k, "none") != "none" else 0)
+        for k, v in depth2.items()
+    )
     x4 = chunk.trim(x3, trim)
     return x4
 
@@ -434,14 +498,22 @@ def add_dummy_padding(x, depth, boundary):
     """
     for k, v in boundary.items():
         d = depth.get(k, 0)
-        if v == 'none' and d > 0:
+        if v == "none" and d > 0:
             empty_shape = list(x.shape)
             empty_shape[k] = d
 
             empty_chunks = list(x.chunks)
             empty_chunks[k] = (d,)
 
-            empty = wrap.empty(empty_shape, chunks=empty_chunks, dtype=x.dtype)
+            try:
+                empty = wrap.empty_like(
+                    getattr(x, "_meta", x),
+                    shape=empty_shape,
+                    chunks=empty_chunks,
+                    dtype=x.dtype,
+                )
+            except TypeError:
+                empty = wrap.empty(empty_shape, chunks=empty_chunks, dtype=x.dtype)
 
             out_chunks = list(x.chunks)
             ax_chunks = list(out_chunks[k])
@@ -466,7 +538,10 @@ def map_overlap(x, func, depth, boundary=None, trim=True, **kwargs):
         The function to apply to each extended block
     depth: int, tuple, or dict
         The number of elements that each block should share with its neighbors
-        If a tuple or dict then this can be different per axis
+        If a tuple or dict then this can be different per axis.
+        Asymmetric depths may be specified using a dict value of (-/+) tuples.
+        Note that asymmetric depths are currently only supported when
+        ``boundary`` is 'none'.
     boundary: str, tuple, dict
         How to handle the boundaries.
         Values include 'reflect', 'periodic', 'nearest', 'none',
@@ -512,8 +587,19 @@ def map_overlap(x, func, depth, boundary=None, trim=True, **kwargs):
     depth2 = coerce_depth(x.ndim, depth)
     boundary2 = coerce_boundary(x.ndim, boundary)
 
+    for i in range(x.ndim):
+        if isinstance(depth2[i], tuple) and boundary2[i] != "none":
+            raise NotImplementedError(
+                "Asymmetric overlap is currently only implemented "
+                "for boundary='none', however boundary for dimension "
+                "{} is {}".format(i, boundary2[i])
+            )
+
+    assert all(type(c) is int for cc in x.chunks for c in cc)
     g = overlap(x, depth=depth2, boundary=boundary2)
+    assert all(type(c) is int for cc in g.chunks for c in cc)
     g2 = g.map_blocks(func, **kwargs)
+    assert all(type(c) is int for cc in g2.chunks for c in cc)
     if trim:
         return trim_internal(g2, depth2, boundary2)
     else:
@@ -525,16 +611,23 @@ def coerce_depth(ndim, depth):
         depth = (depth,) * ndim
     if isinstance(depth, tuple):
         depth = dict(zip(range(ndim), depth))
-
+    if isinstance(depth, dict):
+        for i in range(ndim):
+            if i not in depth:
+                depth.update({i: 0})
     return depth
 
 
 def coerce_boundary(ndim, boundary):
+    default = "reflect"
     if boundary is None:
-        boundary = 'reflect'
+        boundary = default
     if not isinstance(boundary, (tuple, dict)):
         boundary = (boundary,) * ndim
     if isinstance(boundary, tuple):
         boundary = dict(zip(range(ndim), boundary))
-
+    if isinstance(boundary, dict):
+        for i in range(ndim):
+            if i not in boundary:
+                boundary.update({i: default})
     return boundary
