@@ -1,8 +1,7 @@
-from __future__ import absolute_import
-
 import pytest
-pytest.importorskip('numpy')
-pytest.importorskip('scipy')
+
+pytest.importorskip("numpy")
+pytest.importorskip("scipy")
 
 import numpy as np
 import scipy.linalg
@@ -12,29 +11,52 @@ from dask.array.linalg import tsqr, sfqr, svd_compressed, qr, svd
 from dask.array.utils import assert_eq, same_keys
 
 
-@pytest.mark.parametrize('m,n,chunks,error_type', [
-    (20, 10, 10, None),        # tall-skinny regular blocks
-    (20, 10, (3, 10), None),   # tall-skinny regular fat layers
-    (20, 10, ((8, 4, 8), 10), None),   # tall-skinny irregular fat layers
-    (40, 10, ((15, 5, 5, 8, 7), (10)), None),  # tall-skinny non-uniform chunks (why?)
-    (128, 2, (16, 2), None),    # tall-skinny regular thin layers; recursion_depth=1
-    (129, 2, (16, 2), None),    # tall-skinny regular thin layers; recursion_depth=2 --> 17x2
-    (130, 2, (16, 2), None),    # tall-skinny regular thin layers; recursion_depth=2 --> 18x2 next
-    (131, 2, (16, 2), None),    # tall-skinny regular thin layers; recursion_depth=2 --> 18x2 next
-    (300, 10, (40, 10), None),  # tall-skinny regular thin layers; recursion_depth=2
-    (300, 10, (30, 10), None),  # tall-skinny regular thin layers; recursion_depth=3
-    (300, 10, (20, 10), None),  # tall-skinny regular thin layers; recursion_depth=4
-    (10, 5, 10, None),         # single block tall
-    (5, 10, 10, None),         # single block short
-    (10, 10, 10, None),        # single block square
-    (10, 40, (10, 10), ValueError),  # short-fat regular blocks
-    (10, 40, (10, 15), ValueError),  # short-fat irregular blocks
-    (10, 40, ((10), (15, 5, 5, 8, 7)), ValueError),  # short-fat non-uniform chunks (why?)
-    (20, 20, 10, ValueError),  # 2x2 regular blocks
-])
+@pytest.mark.parametrize(
+    "m,n,chunks,error_type",
+    [
+        (20, 10, 10, None),  # tall-skinny regular blocks
+        (20, 10, (3, 10), None),  # tall-skinny regular fat layers
+        (20, 10, ((8, 4, 8), 10), None),  # tall-skinny irregular fat layers
+        (40, 10, ((15, 5, 5, 8, 7), 10), None),  # tall-skinny non-uniform chunks (why?)
+        (128, 2, (16, 2), None),  # tall-skinny regular thin layers; recursion_depth=1
+        (
+            129,
+            2,
+            (16, 2),
+            None,
+        ),  # tall-skinny regular thin layers; recursion_depth=2 --> 17x2
+        (
+            130,
+            2,
+            (16, 2),
+            None,
+        ),  # tall-skinny regular thin layers; recursion_depth=2 --> 18x2 next
+        (
+            131,
+            2,
+            (16, 2),
+            None,
+        ),  # tall-skinny regular thin layers; recursion_depth=2 --> 18x2 next
+        (300, 10, (40, 10), None),  # tall-skinny regular thin layers; recursion_depth=2
+        (300, 10, (30, 10), None),  # tall-skinny regular thin layers; recursion_depth=3
+        (300, 10, (20, 10), None),  # tall-skinny regular thin layers; recursion_depth=4
+        (10, 5, 10, None),  # single block tall
+        (5, 10, 10, None),  # single block short
+        (10, 10, 10, None),  # single block square
+        (10, 40, (10, 10), ValueError),  # short-fat regular blocks
+        (10, 40, (10, 15), ValueError),  # short-fat irregular blocks
+        (
+            10,
+            40,
+            (10, (15, 5, 5, 8, 7)),
+            ValueError,
+        ),  # short-fat non-uniform chunks (why?)
+        (20, 20, 10, ValueError),  # 2x2 regular blocks
+    ],
+)
 def test_tsqr(m, n, chunks, error_type):
     mat = np.random.rand(m, n)
-    data = da.from_array(mat, chunks=chunks, name='A')
+    data = da.from_array(mat, chunks=chunks, name="A")
 
     # qr
     m_q = m
@@ -76,32 +98,98 @@ def test_tsqr(m, n, chunks, error_type):
             u, s, vh = tsqr(data, compute_svd=True)
 
 
-@pytest.mark.parametrize('m_min,n_max,chunks,vary_rows,vary_cols,error_type', [
-    (10, 5, (10, 5), True, False, None),   # single block tall
-    (10, 5, (10, 5), False, True, None),   # single block tall
-    (10, 5, (10, 5), True, True, None),    # single block tall
-    (40, 5, (10, 5), True, False, None),   # multiple blocks tall
-    (40, 5, (10, 5), False, True, None),   # multiple blocks tall
-    (40, 5, (10, 5), True, True, None),    # multiple blocks tall
-    (300, 10, (40, 10), True, False, None),  # tall-skinny regular thin layers; recursion_depth=2
-    (300, 10, (30, 10), True, False, None),  # tall-skinny regular thin layers; recursion_depth=3
-    (300, 10, (20, 10), True, False, None),  # tall-skinny regular thin layers; recursion_depth=4
-    (300, 10, (40, 10), False, True, None),  # tall-skinny regular thin layers; recursion_depth=2
-    (300, 10, (30, 10), False, True, None),  # tall-skinny regular thin layers; recursion_depth=3
-    (300, 10, (20, 10), False, True, None),  # tall-skinny regular thin layers; recursion_depth=4
-    (300, 10, (40, 10), True, True, None),  # tall-skinny regular thin layers; recursion_depth=2
-    (300, 10, (30, 10), True, True, None),  # tall-skinny regular thin layers; recursion_depth=3
-    (300, 10, (20, 10), True, True, None),  # tall-skinny regular thin layers; recursion_depth=4
-])
+@pytest.mark.parametrize(
+    "m_min,n_max,chunks,vary_rows,vary_cols,error_type",
+    [
+        (10, 5, (10, 5), True, False, None),  # single block tall
+        (10, 5, (10, 5), False, True, None),  # single block tall
+        (10, 5, (10, 5), True, True, None),  # single block tall
+        (40, 5, (10, 5), True, False, None),  # multiple blocks tall
+        (40, 5, (10, 5), False, True, None),  # multiple blocks tall
+        (40, 5, (10, 5), True, True, None),  # multiple blocks tall
+        (
+            300,
+            10,
+            (40, 10),
+            True,
+            False,
+            None,
+        ),  # tall-skinny regular thin layers; recursion_depth=2
+        (
+            300,
+            10,
+            (30, 10),
+            True,
+            False,
+            None,
+        ),  # tall-skinny regular thin layers; recursion_depth=3
+        (
+            300,
+            10,
+            (20, 10),
+            True,
+            False,
+            None,
+        ),  # tall-skinny regular thin layers; recursion_depth=4
+        (
+            300,
+            10,
+            (40, 10),
+            False,
+            True,
+            None,
+        ),  # tall-skinny regular thin layers; recursion_depth=2
+        (
+            300,
+            10,
+            (30, 10),
+            False,
+            True,
+            None,
+        ),  # tall-skinny regular thin layers; recursion_depth=3
+        (
+            300,
+            10,
+            (20, 10),
+            False,
+            True,
+            None,
+        ),  # tall-skinny regular thin layers; recursion_depth=4
+        (
+            300,
+            10,
+            (40, 10),
+            True,
+            True,
+            None,
+        ),  # tall-skinny regular thin layers; recursion_depth=2
+        (
+            300,
+            10,
+            (30, 10),
+            True,
+            True,
+            None,
+        ),  # tall-skinny regular thin layers; recursion_depth=3
+        (
+            300,
+            10,
+            (20, 10),
+            True,
+            True,
+            None,
+        ),  # tall-skinny regular thin layers; recursion_depth=4
+    ],
+)
 def test_tsqr_uncertain(m_min, n_max, chunks, vary_rows, vary_cols, error_type):
     mat = np.random.rand(m_min * 2, n_max)
     m, n = m_min * 2, n_max
     mat[0:m_min, 0] += 1
     _c0 = mat[:, 0]
     _r0 = mat[0, :]
-    c0 = da.from_array(_c0, chunks=m_min, name='c')
-    r0 = da.from_array(_r0, chunks=n_max, name='r')
-    data = da.from_array(mat, chunks=chunks, name='A')
+    c0 = da.from_array(_c0, chunks=m_min, name="c")
+    r0 = da.from_array(_r0, chunks=n_max, name="r")
+    data = da.from_array(mat, chunks=chunks, name="A")
     if vary_rows:
         data = data[c0 > 0.5, :]
         mat = mat[_c0 > 0.5, :]
@@ -173,7 +261,7 @@ def test_tsqr_zero_height_chunks():
     assert_eq(r, da.triu(r.rechunk(r.shape[0])))  # r must be upper triangular
 
     # uncertainty
-    mat2 = np.vstack([mat, -np.ones((10, 5))])
+    mat2 = np.vstack([mat, -(np.ones((10, 5)))])
     v2 = mat2[:, 0]
     x2 = da.from_array(mat2, chunks=5)
     c = da.from_array(v2, chunks=5)
@@ -188,29 +276,72 @@ def test_tsqr_zero_height_chunks():
     assert_eq(r, np.triu(r))  # r must be upper triangular
 
 
-@pytest.mark.parametrize('m,n,chunks,error_type', [
-    (20, 10, 10, ValueError),        # tall-skinny regular blocks
-    (20, 10, (3, 10), ValueError),   # tall-skinny regular fat layers
-    (20, 10, ((8, 4, 8), 10), ValueError),   # tall-skinny irregular fat layers
-    (40, 10, ((15, 5, 5, 8, 7), (10)), ValueError),  # tall-skinny non-uniform chunks (why?)
-    (128, 2, (16, 2), ValueError),    # tall-skinny regular thin layers; recursion_depth=1
-    (129, 2, (16, 2), ValueError),    # tall-skinny regular thin layers; recursion_depth=2 --> 17x2
-    (130, 2, (16, 2), ValueError),    # tall-skinny regular thin layers; recursion_depth=2 --> 18x2 next
-    (131, 2, (16, 2), ValueError),    # tall-skinny regular thin layers; recursion_depth=2 --> 18x2 next
-    (300, 10, (40, 10), ValueError),  # tall-skinny regular thin layers; recursion_depth=2
-    (300, 10, (30, 10), ValueError),  # tall-skinny regular thin layers; recursion_depth=3
-    (300, 10, (20, 10), ValueError),  # tall-skinny regular thin layers; recursion_depth=4
-    (10, 5, 10, None),         # single block tall
-    (5, 10, 10, None),         # single block short
-    (10, 10, 10, None),        # single block square
-    (10, 40, (10, 10), None),  # short-fat regular blocks
-    (10, 40, (10, 15), None),  # short-fat irregular blocks
-    (10, 40, ((10), (15, 5, 5, 8, 7)), None),  # short-fat non-uniform chunks (why?)
-    (20, 20, 10, ValueError),  # 2x2 regular blocks
-])
+@pytest.mark.parametrize(
+    "m,n,chunks,error_type",
+    [
+        (20, 10, 10, ValueError),  # tall-skinny regular blocks
+        (20, 10, (3, 10), ValueError),  # tall-skinny regular fat layers
+        (20, 10, ((8, 4, 8), 10), ValueError),  # tall-skinny irregular fat layers
+        (
+            40,
+            10,
+            ((15, 5, 5, 8, 7), 10),
+            ValueError,
+        ),  # tall-skinny non-uniform chunks (why?)
+        (
+            128,
+            2,
+            (16, 2),
+            ValueError,
+        ),  # tall-skinny regular thin layers; recursion_depth=1
+        (
+            129,
+            2,
+            (16, 2),
+            ValueError,
+        ),  # tall-skinny regular thin layers; recursion_depth=2 --> 17x2
+        (
+            130,
+            2,
+            (16, 2),
+            ValueError,
+        ),  # tall-skinny regular thin layers; recursion_depth=2 --> 18x2 next
+        (
+            131,
+            2,
+            (16, 2),
+            ValueError,
+        ),  # tall-skinny regular thin layers; recursion_depth=2 --> 18x2 next
+        (
+            300,
+            10,
+            (40, 10),
+            ValueError,
+        ),  # tall-skinny regular thin layers; recursion_depth=2
+        (
+            300,
+            10,
+            (30, 10),
+            ValueError,
+        ),  # tall-skinny regular thin layers; recursion_depth=3
+        (
+            300,
+            10,
+            (20, 10),
+            ValueError,
+        ),  # tall-skinny regular thin layers; recursion_depth=4
+        (10, 5, 10, None),  # single block tall
+        (5, 10, 10, None),  # single block short
+        (10, 10, 10, None),  # single block square
+        (10, 40, (10, 10), None),  # short-fat regular blocks
+        (10, 40, (10, 15), None),  # short-fat irregular blocks
+        (10, 40, (10, (15, 5, 5, 8, 7)), None),  # short-fat non-uniform chunks (why?)
+        (20, 20, 10, ValueError),  # 2x2 regular blocks
+    ],
+)
 def test_sfqr(m, n, chunks, error_type):
     mat = np.random.rand(m, n)
-    data = da.from_array(mat, chunks=chunks, name='A')
+    data = da.from_array(mat, chunks=chunks, name="A")
     m_q = m
     n_q = min(m, n)
     m_r = n_q
@@ -229,29 +360,47 @@ def test_sfqr(m, n, chunks, error_type):
             q, r = sfqr(data)
 
 
-@pytest.mark.parametrize('m,n,chunks,error_type', [
-    (20, 10, 10, None),        # tall-skinny regular blocks
-    (20, 10, (3, 10), None),   # tall-skinny regular fat layers
-    (20, 10, ((8, 4, 8), 10), None),   # tall-skinny irregular fat layers
-    (40, 10, ((15, 5, 5, 8, 7), (10)), None),  # tall-skinny non-uniform chunks (why?)
-    (128, 2, (16, 2), None),    # tall-skinny regular thin layers; recursion_depth=1
-    (129, 2, (16, 2), None),    # tall-skinny regular thin layers; recursion_depth=2 --> 17x2
-    (130, 2, (16, 2), None),    # tall-skinny regular thin layers; recursion_depth=2 --> 18x2 next
-    (131, 2, (16, 2), None),    # tall-skinny regular thin layers; recursion_depth=2 --> 18x2 next
-    (300, 10, (40, 10), None),  # tall-skinny regular thin layers; recursion_depth=2
-    (300, 10, (30, 10), None),  # tall-skinny regular thin layers; recursion_depth=3
-    (300, 10, (20, 10), None),  # tall-skinny regular thin layers; recursion_depth=4
-    (10, 5, 10, None),         # single block tall
-    (5, 10, 10, None),         # single block short
-    (10, 10, 10, None),        # single block square
-    (10, 40, (10, 10), None),  # short-fat regular blocks
-    (10, 40, (10, 15), None),  # short-fat irregular blocks
-    (10, 40, ((10), (15, 5, 5, 8, 7)), None),  # short-fat non-uniform chunks (why?)
-    (20, 20, 10, NotImplementedError),  # 2x2 regular blocks
-])
+@pytest.mark.parametrize(
+    "m,n,chunks,error_type",
+    [
+        (20, 10, 10, None),  # tall-skinny regular blocks
+        (20, 10, (3, 10), None),  # tall-skinny regular fat layers
+        (20, 10, ((8, 4, 8), 10), None),  # tall-skinny irregular fat layers
+        (40, 10, ((15, 5, 5, 8, 7), 10), None),  # tall-skinny non-uniform chunks (why?)
+        (128, 2, (16, 2), None),  # tall-skinny regular thin layers; recursion_depth=1
+        (
+            129,
+            2,
+            (16, 2),
+            None,
+        ),  # tall-skinny regular thin layers; recursion_depth=2 --> 17x2
+        (
+            130,
+            2,
+            (16, 2),
+            None,
+        ),  # tall-skinny regular thin layers; recursion_depth=2 --> 18x2 next
+        (
+            131,
+            2,
+            (16, 2),
+            None,
+        ),  # tall-skinny regular thin layers; recursion_depth=2 --> 18x2 next
+        (300, 10, (40, 10), None),  # tall-skinny regular thin layers; recursion_depth=2
+        (300, 10, (30, 10), None),  # tall-skinny regular thin layers; recursion_depth=3
+        (300, 10, (20, 10), None),  # tall-skinny regular thin layers; recursion_depth=4
+        (10, 5, 10, None),  # single block tall
+        (5, 10, 10, None),  # single block short
+        (10, 10, 10, None),  # single block square
+        (10, 40, (10, 10), None),  # short-fat regular blocks
+        (10, 40, (10, 15), None),  # short-fat irregular blocks
+        (10, 40, (10, (15, 5, 5, 8, 7)), None),  # short-fat non-uniform chunks (why?)
+        (20, 20, 10, NotImplementedError),  # 2x2 regular blocks
+    ],
+)
 def test_qr(m, n, chunks, error_type):
     mat = np.random.rand(m, n)
-    data = da.from_array(mat, chunks=chunks, name='A')
+    data = da.from_array(mat, chunks=chunks, name="A")
     m_q = m
     n_q = min(m, n)
     m_r = n_q
@@ -273,7 +422,7 @@ def test_qr(m, n, chunks, error_type):
 def test_linalg_consistent_names():
     m, n = 20, 10
     mat = np.random.rand(m, n)
-    data = da.from_array(mat, chunks=(10, n), name='A')
+    data = da.from_array(mat, chunks=(10, n), name="A")
 
     q1, r1 = qr(data)
     q2, r2 = qr(data)
@@ -287,14 +436,10 @@ def test_linalg_consistent_names():
     assert same_keys(v1, v2)
 
 
-@pytest.mark.parametrize("m,n", [
-    (10, 20),
-    (15, 15),
-    (20, 10),
-])
+@pytest.mark.parametrize("m,n", [(10, 20), (15, 15), (20, 10)])
 def test_dask_svd_self_consistent(m, n):
     a = np.random.rand(m, n)
-    d_a = da.from_array(a, chunks=(3, n), name='A')
+    d_a = da.from_array(a, chunks=(3, n), name="A")
 
     d_u, d_s, d_vt = da.linalg.svd(d_a)
     u, s, vt = da.compute(d_u, d_s, d_vt)
@@ -319,9 +464,9 @@ def test_svd_compressed():
     usvt = da.dot(u, da.dot(da.diag(s), vt))
 
     tol = 0.2
-    assert_eq(da.linalg.norm(usvt),
-              np.linalg.norm(mat),
-              rtol=tol, atol=tol)  # average accuracy check
+    assert_eq(
+        da.linalg.norm(usvt), np.linalg.norm(mat), rtol=tol, atol=tol
+    )  # average accuracy check
 
     u = u[:, :r]
     s = s[:r]
@@ -353,15 +498,18 @@ def _check_lu_result(p, l, u, A):
 
 
 def test_lu_1():
-    A1 = np.array([[7, 3, -1, 2], [3, 8, 1, -4],
-                  [-1, 1, 4, -1], [2, -4, -1, 6] ])
+    A1 = np.array([[7, 3, -1, 2], [3, 8, 1, -4], [-1, 1, 4, -1], [2, -4, -1, 6]])
 
-    A2 = np.array([[7,  0,  0,  0,  0,  0],
-                   [0,  8,  0,  0,  0,  0],
-                   [0,  0,  4,  0,  0,  0],
-                   [0,  0,  0,  6,  0,  0],
-                   [0,  0,  0,  0,  3,  0],
-                   [0,  0,  0,  0,  0,  5]])
+    A2 = np.array(
+        [
+            [7, 0, 0, 0, 0, 0],
+            [0, 8, 0, 0, 0, 0],
+            [0, 0, 4, 0, 0, 0],
+            [0, 0, 0, 6, 0, 0],
+            [0, 0, 0, 0, 3, 0],
+            [0, 0, 0, 0, 0, 5],
+        ]
+    )
     # without shuffle
     for A, chunk in zip([A1, A2], [2, 2]):
         dA = da.from_array(A, chunks=(chunk, chunk))
@@ -372,12 +520,16 @@ def test_lu_1():
         assert_eq(u, du, check_graph=False)
         _check_lu_result(dp, dl, du, A)
 
-    A3 = np.array([[ 7,  3,  2,  1,  4,  1],
-                   [ 7, 11,  5,  2,  5,  2],
-                   [21, 25, 16, 10, 16,  5],
-                   [21, 41, 18, 13, 16, 11],
-                   [14, 46, 23, 24, 21, 22],
-                   [ 0, 56, 29, 17, 14, 8]])
+    A3 = np.array(
+        [
+            [7, 3, 2, 1, 4, 1],
+            [7, 11, 5, 2, 5, 2],
+            [21, 25, 16, 10, 16, 5],
+            [21, 41, 18, 13, 16, 11],
+            [14, 46, 23, 24, 21, 22],
+            [0, 56, 29, 17, 14, 8],
+        ]
+    )
 
     # with shuffle
     for A, chunk in zip([A3], [2]):
@@ -388,7 +540,8 @@ def test_lu_1():
 
 
 @pytest.mark.slow
-@pytest.mark.parametrize('size', [10, 20, 30, 50])
+@pytest.mark.parametrize("size", [10, 20, 30, 50])
+@pytest.mark.filterwarnings("ignore:Increasing:dask.array.core.PerformanceWarning")
 def test_lu_2(size):
     np.random.seed(10)
     A = np.random.randint(0, 10, (size, size))
@@ -399,7 +552,7 @@ def test_lu_2(size):
 
 
 @pytest.mark.slow
-@pytest.mark.parametrize('size', [50, 100, 200])
+@pytest.mark.parametrize("size", [50, 100, 200])
 def test_lu_3(size):
     np.random.seed(10)
     A = np.random.randint(0, 10, (size, size))
@@ -423,7 +576,7 @@ def test_lu_errors():
     pytest.raises(ValueError, lambda: da.linalg.lu(dA))
 
 
-@pytest.mark.parametrize(('shape', 'chunk'), [(20, 10), (50, 10), (70, 20)])
+@pytest.mark.parametrize(("shape", "chunk"), [(20, 10), (50, 10), (70, 20)])
 def test_solve_triangular_vector(shape, chunk):
     np.random.seed(1)
 
@@ -447,7 +600,7 @@ def test_solve_triangular_vector(shape, chunk):
     assert_eq(dAl.dot(res), b.astype(float))
 
 
-@pytest.mark.parametrize(('shape', 'chunk'), [(20, 10), (50, 10), (50, 20)])
+@pytest.mark.parametrize(("shape", "chunk"), [(20, 10), (50, 10), (50, 20)])
 def test_solve_triangular_matrix(shape, chunk):
     np.random.seed(1)
 
@@ -471,7 +624,7 @@ def test_solve_triangular_matrix(shape, chunk):
     assert_eq(dAl.dot(res), b.astype(float))
 
 
-@pytest.mark.parametrize(('shape', 'chunk'), [(20, 10), (50, 10), (50, 20)])
+@pytest.mark.parametrize(("shape", "chunk"), [(20, 10), (50, 10), (50, 20)])
 def test_solve_triangular_matrix2(shape, chunk):
     np.random.seed(1)
 
@@ -509,7 +662,7 @@ def test_solve_triangular_errors():
     pytest.raises(ValueError, lambda: da.linalg.solve_triangular(dA, db))
 
 
-@pytest.mark.parametrize(('shape', 'chunk'), [(20, 10), (50, 10)])
+@pytest.mark.parametrize(("shape", "chunk"), [(20, 10), (50, 10)])
 def test_solve(shape, chunk):
     np.random.seed(1)
 
@@ -541,7 +694,7 @@ def test_solve(shape, chunk):
     assert_eq(dA.dot(res), b.astype(float), check_graph=False)
 
 
-@pytest.mark.parametrize(('shape', 'chunk'), [(20, 10), (50, 10)])
+@pytest.mark.parametrize(("shape", "chunk"), [(20, 10), (50, 10)])
 def test_inv(shape, chunk):
     np.random.seed(1)
 
@@ -560,7 +713,7 @@ def _get_symmat(size):
     return lA.dot(lA.T)
 
 
-@pytest.mark.parametrize(('shape', 'chunk'), [(20, 10), (30, 6)])
+@pytest.mark.parametrize(("shape", "chunk"), [(20, 10), (30, 6)])
 def test_solve_sym_pos(shape, chunk):
     np.random.seed(1)
 
@@ -592,17 +745,20 @@ def test_solve_sym_pos(shape, chunk):
     assert_eq(dA.dot(res), b.astype(float), check_graph=False)
 
 
-@pytest.mark.parametrize(('shape', 'chunk'), [(20, 10), (12, 3), (30, 3), (30, 6)])
+@pytest.mark.parametrize(("shape", "chunk"), [(20, 10), (12, 3), (30, 3), (30, 6)])
 def test_cholesky(shape, chunk):
 
     A = _get_symmat(shape)
     dA = da.from_array(A, (chunk, chunk))
     assert_eq(da.linalg.cholesky(dA), scipy.linalg.cholesky(A), check_graph=False)
-    assert_eq(da.linalg.cholesky(dA, lower=True), scipy.linalg.cholesky(A, lower=True), check_graph=False)
+    assert_eq(
+        da.linalg.cholesky(dA, lower=True),
+        scipy.linalg.cholesky(A, lower=True),
+        check_graph=False,
+    )
 
 
-@pytest.mark.parametrize(("nrow", "ncol", "chunk"),
-                         [(20, 10, 5), (100, 10, 10)])
+@pytest.mark.parametrize(("nrow", "ncol", "chunk"), [(20, 10, 5), (100, 10, 10)])
 def test_lstsq(nrow, ncol, chunk):
     np.random.seed(1)
     A = np.random.randint(1, 20, (nrow, ncol))
@@ -611,7 +767,7 @@ def test_lstsq(nrow, ncol, chunk):
     dA = da.from_array(A, (chunk, ncol))
     db = da.from_array(b, chunk)
 
-    x, r, rank, s = np.linalg.lstsq(A, b)
+    x, r, rank, s = np.linalg.lstsq(A, b, rcond=-1)
     dx, dr, drank, ds = da.linalg.lstsq(dA, db)
 
     assert_eq(dx, x)
@@ -623,9 +779,9 @@ def test_lstsq(nrow, ncol, chunk):
     A[:, 1] = A[:, 2]
     dA = da.from_array(A, (chunk, ncol))
     db = da.from_array(b, chunk)
-    x, r, rank, s = np.linalg.lstsq(A, b,
-                                    rcond=np.finfo(np.double).eps * max(nrow,
-                                                                        ncol))
+    x, r, rank, s = np.linalg.lstsq(
+        A, b, rcond=np.finfo(np.double).eps * max(nrow, ncol)
+    )
     assert rank == ncol - 1
     dx, dr, drank, ds = da.linalg.lstsq(dA, db)
     assert drank.compute() == rank
@@ -635,16 +791,14 @@ def test_no_chunks_svd():
     x = np.random.random((100, 10))
     u, s, v = np.linalg.svd(x, full_matrices=0)
 
-    for chunks in [((np.nan,) * 10, (10,)),
-                   ((np.nan,) * 10, (np.nan,))]:
+    for chunks in [((np.nan,) * 10, (10,)), ((np.nan,) * 10, (np.nan,))]:
         dx = da.from_array(x, chunks=(10, 10))
         dx._chunks = chunks
 
         du, ds, dv = da.linalg.svd(dx)
 
         assert_eq(s, ds)
-        assert_eq(u.dot(np.diag(s)).dot(v),
-                  du.dot(da.diag(ds)).dot(dv))
+        assert_eq(u.dot(np.diag(s)).dot(v), du.dot(da.diag(ds)).dot(dv))
         assert_eq(du.T.dot(du), np.eye(10))
         assert_eq(dv.T.dot(dv), np.eye(10))
 
@@ -654,23 +808,12 @@ def test_no_chunks_svd():
         assert_eq(abs(u), abs(du))
 
 
-@pytest.mark.parametrize("shape, chunks, axis", [
-    [(5,), (2,), None],
-    [(5,), (2,), 0],
-    [(5,), (2,), (0,)],
-    [(5, 6), (2, 2), None],
-])
-@pytest.mark.parametrize("norm", [
-    None,
-    1,
-    -1,
-    np.inf,
-    -np.inf,
-])
-@pytest.mark.parametrize("keepdims", [
-    False,
-    True,
-])
+@pytest.mark.parametrize(
+    "shape, chunks, axis",
+    [[(5,), (2,), None], [(5,), (2,), 0], [(5,), (2,), (0,)], [(5, 6), (2, 2), None]],
+)
+@pytest.mark.parametrize("norm", [None, 1, -1, np.inf, -np.inf])
+@pytest.mark.parametrize("keepdims", [False, True])
 def test_norm_any_ndim(shape, chunks, axis, norm, keepdims):
     a = np.random.random(shape)
     d = da.from_array(a, chunks=chunks)
@@ -682,24 +825,18 @@ def test_norm_any_ndim(shape, chunks, axis, norm, keepdims):
 
 
 @pytest.mark.slow
-@pytest.mark.parametrize("shape, chunks", [
-    [(5,), (2,)],
-    [(5, 3), (2, 2)],
-    [(4, 5, 3), (2, 2, 2)],
-    [(4, 5, 2, 3), (2, 2, 2, 2)],
-    [(2, 5, 2, 4, 3), (2, 2, 2, 2, 2)],
-])
-@pytest.mark.parametrize("norm", [
-    None,
-    1,
-    -1,
-    np.inf,
-    -np.inf,
-])
-@pytest.mark.parametrize("keepdims", [
-    False,
-    True,
-])
+@pytest.mark.parametrize(
+    "shape, chunks",
+    [
+        [(5,), (2,)],
+        [(5, 3), (2, 2)],
+        [(4, 5, 3), (2, 2, 2)],
+        [(4, 5, 2, 3), (2, 2, 2, 2)],
+        [(2, 5, 2, 4, 3), (2, 2, 2, 2, 2)],
+    ],
+)
+@pytest.mark.parametrize("norm", [None, 1, -1, np.inf, -np.inf])
+@pytest.mark.parametrize("keepdims", [False, True])
 def test_norm_any_slice(shape, chunks, norm, keepdims):
     a = np.random.random(shape)
     d = da.from_array(a, chunks=chunks)
@@ -715,21 +852,11 @@ def test_norm_any_slice(shape, chunks, norm, keepdims):
             assert_eq(a_r, d_r)
 
 
-@pytest.mark.parametrize("shape, chunks, axis", [
-    [(5,), (2,), None],
-    [(5,), (2,), 0],
-    [(5,), (2,), (0,)],
-])
-@pytest.mark.parametrize("norm", [
-    0,
-    2,
-    -2,
-    0.5,
-])
-@pytest.mark.parametrize("keepdims", [
-    False,
-    True,
-])
+@pytest.mark.parametrize(
+    "shape, chunks, axis", [[(5,), (2,), None], [(5,), (2,), 0], [(5,), (2,), (0,)]]
+)
+@pytest.mark.parametrize("norm", [0, 2, -2, 0.5])
+@pytest.mark.parametrize("keepdims", [False, True])
 def test_norm_1dim(shape, chunks, axis, norm, keepdims):
     a = np.random.random(shape)
     d = da.from_array(a, chunks=chunks)
@@ -739,21 +866,12 @@ def test_norm_1dim(shape, chunks, axis, norm, keepdims):
     assert_eq(a_r, d_r)
 
 
-@pytest.mark.parametrize("shape, chunks, axis", [
-    [(5, 6), (2, 2), None],
-    [(5, 6), (2, 2), (0, 1)],
-    [(5, 6), (2, 2), (1, 0)],
-])
-@pytest.mark.parametrize("norm", [
-    "fro",
-    "nuc",
-    2,
-    -2
-])
-@pytest.mark.parametrize("keepdims", [
-    False,
-    True,
-])
+@pytest.mark.parametrize(
+    "shape, chunks, axis",
+    [[(5, 6), (2, 2), None], [(5, 6), (2, 2), (0, 1)], [(5, 6), (2, 2), (1, 0)]],
+)
+@pytest.mark.parametrize("norm", ["fro", "nuc", 2, -2])
+@pytest.mark.parametrize("keepdims", [False, True])
 def test_norm_2dim(shape, chunks, axis, norm, keepdims):
     a = np.random.random(shape)
     d = da.from_array(a, chunks=chunks)
@@ -768,22 +886,25 @@ def test_norm_2dim(shape, chunks, axis, norm, keepdims):
     assert_eq(a_r, d_r)
 
 
-@pytest.mark.parametrize("shape, chunks, axis", [
-    [(3, 2, 4), (2, 2, 2), (1, 2)],
-    [(2, 3, 4, 5), (2, 2, 2, 2), (-1, -2)],
-])
-@pytest.mark.parametrize("norm", [
-    "nuc",
-    2,
-    -2
-])
-@pytest.mark.parametrize("keepdims", [
-    False,
-    True,
-])
+@pytest.mark.parametrize(
+    "shape, chunks, axis",
+    [[(3, 2, 4), (2, 2, 2), (1, 2)], [(2, 3, 4, 5), (2, 2, 2, 2), (-1, -2)]],
+)
+@pytest.mark.parametrize("norm", ["nuc", 2, -2])
+@pytest.mark.parametrize("keepdims", [False, True])
 def test_norm_implemented_errors(shape, chunks, axis, norm, keepdims):
     a = np.random.random(shape)
     d = da.from_array(a, chunks=chunks)
     if len(shape) > 2 and len(axis) == 2:
         with pytest.raises(NotImplementedError):
             da.linalg.norm(d, ord=norm, axis=axis, keepdims=keepdims)
+
+
+def test_svd_compressed_compute():
+    x = da.ones((100, 100), chunks=(10, 10))
+    u, s, v = da.linalg.svd_compressed(x, k=2, n_power_iter=0, compute=True, seed=123)
+    uu, ss, vv = da.linalg.svd_compressed(x, k=2, n_power_iter=0, seed=123)
+
+    assert len(v.dask) < len(vv.dask)
+
+    assert_eq(v, vv)
