@@ -1,16 +1,16 @@
-from __future__ import absolute_import, division, print_function
-
 from datetime import datetime
 from collections import defaultdict
 
-from toolz import merge
 import bisect
 import numpy as np
 import pandas as pd
 
 from .core import new_dd_object, Series
+from ..array.core import Array
+from .utils import is_index_like, meta_nonempty
 from . import methods
 from ..base import tokenize
+from ..highlevelgraph import HighLevelGraph
 
 
 class _IndexerBase(object):
@@ -36,7 +36,6 @@ class _IndexerBase(object):
 
 
 class _iLocIndexer(_IndexerBase):
-
     @property
     def _meta_indexer(self):
         return self.obj._meta.iloc
@@ -44,24 +43,26 @@ class _iLocIndexer(_IndexerBase):
     def __getitem__(self, key):
 
         # dataframe
-        msg = ("'DataFrame.iloc' only supports selecting columns. "
-               "It must be used like 'df.iloc[:, column_indexer]'.")
+        msg = (
+            "'DataFrame.iloc' only supports selecting columns. "
+            "It must be used like 'df.iloc[:, column_indexer]'."
+        )
         if not isinstance(key, tuple):
-            raise ValueError(msg)
+            raise NotImplementedError(msg)
 
-        if len(key) != 2:
-            raise ValueError(msg)
+        if len(key) > 2:
+            raise ValueError("Too many indexers")
 
         iindexer, cindexer = key
 
         if iindexer != slice(None):
-            raise ValueError(msg)
+            raise NotImplementedError(msg)
 
         return self._iloc(iindexer, cindexer)
 
     def _iloc(self, iindexer, cindexer):
         assert iindexer == slice(None)
-        meta = self._make_meta( iindexer, cindexer)
+        meta = self._make_meta(iindexer, cindexer)
 
         return self.obj.map_partitions(methods.iloc, cindexer, meta=meta)
 
@@ -79,7 +80,7 @@ class _LocIndexer(_IndexerBase):
             # multi-dimensional selection
             if len(key) > self.obj.ndim:
                 # raise from pandas
-                msg = 'Too many indexers'
+                msg = "Too many indexers"
                 raise pd.core.indexing.IndexingError(msg)
 
             iindexer = key[0]
@@ -94,6 +95,8 @@ class _LocIndexer(_IndexerBase):
         """ Helper function for the .loc accessor """
         if isinstance(iindexer, Series):
             return self._loc_series(iindexer, cindexer)
+        elif isinstance(iindexer, Array):
+            return self._loc_array(iindexer, cindexer)
 
         if self.obj.known_divisions:
             iindexer = self._maybe_partial_time_string(iindexer)
@@ -109,31 +112,37 @@ class _LocIndexer(_IndexerBase):
             if isinstance(iindexer, (list, np.ndarray)):
                 # applying map_pattition to each partitions
                 # results in duplicated NaN rows
-                msg = 'Cannot index with list against unknown division'
+                msg = "Cannot index with list against unknown division"
                 raise KeyError(msg)
             elif not isinstance(iindexer, slice):
                 iindexer = slice(iindexer, iindexer)
 
             meta = self._make_meta(iindexer, cindexer)
-            return self.obj.map_partitions(methods.try_loc, iindexer, cindexer,
-                                           meta=meta)
+            return self.obj.map_partitions(
+                methods.try_loc, iindexer, cindexer, meta=meta
+            )
 
     def _maybe_partial_time_string(self, iindexer):
         """
         Convert index-indexer for partial time string slicing
         if obj.index is DatetimeIndex / PeriodIndex
         """
-        iindexer = _maybe_partial_time_string(self.obj._meta_nonempty.index,
-                                              iindexer, kind='loc')
+        idx = meta_nonempty(self.obj._meta.index)
+        iindexer = _maybe_partial_time_string(idx, iindexer, kind="loc")
         return iindexer
 
     def _loc_series(self, iindexer, cindexer):
         meta = self._make_meta(iindexer, cindexer)
-        return self.obj.map_partitions(methods.loc, iindexer, cindexer,
-                                       token='loc-series', meta=meta)
+        return self.obj.map_partitions(
+            methods.loc, iindexer, cindexer, token="loc-series", meta=meta
+        )
+
+    def _loc_array(self, iindexer, cindexer):
+        iindexer_series = iindexer.to_dask_dataframe("_", self.obj.index)
+        return self._loc_series(iindexer_series, cindexer)
 
     def _loc_list(self, iindexer, cindexer):
-        name = 'loc-%s' % tokenize(iindexer, self.obj)
+        name = "loc-%s" % tokenize(iindexer, self.obj)
         parts = self._get_partitions(iindexer)
         meta = self._make_meta(iindexer, cindexer)
 
@@ -142,8 +151,7 @@ class _LocIndexer(_IndexerBase):
             divisions = []
             items = sorted(parts.items())
             for i, (div, indexer) in enumerate(items):
-                dsk[name, i] = (methods.loc, (self._name, div),
-                                indexer, cindexer)
+                dsk[name, i] = (methods.loc, (self._name, div), indexer, cindexer)
                 # append minimum value as division
                 divisions.append(sorted(indexer)[0])
             # append maximum value of the last division
@@ -151,22 +159,28 @@ class _LocIndexer(_IndexerBase):
         else:
             divisions = [None, None]
             dsk = {(name, 0): meta.head(0)}
-        return new_dd_object(merge(self.obj.dask, dsk), name,
-                             meta=meta, divisions=divisions)
+        graph = HighLevelGraph.from_collections(name, dsk, dependencies=[self.obj])
+        return new_dd_object(graph, name, meta=meta, divisions=divisions)
 
     def _loc_element(self, iindexer, cindexer):
-        name = 'loc-%s' % tokenize(iindexer, self.obj)
+        name = "loc-%s" % tokenize(iindexer, self.obj)
         part = self._get_partitions(iindexer)
 
         if iindexer < self.obj.divisions[0] or iindexer > self.obj.divisions[-1]:
-            raise KeyError('the label [%s] is not in the index' % str(iindexer))
+            raise KeyError("the label [%s] is not in the index" % str(iindexer))
 
-        dsk = {(name, 0): (methods.loc, (self._name, part),
-                           slice(iindexer, iindexer), cindexer)}
+        dsk = {
+            (name, 0): (
+                methods.loc,
+                (self._name, part),
+                slice(iindexer, iindexer),
+                cindexer,
+            )
+        }
 
         meta = self._make_meta(iindexer, cindexer)
-        return new_dd_object(merge(self.obj.dask, dsk), name,
-                             meta=meta, divisions=[iindexer, iindexer])
+        graph = HighLevelGraph.from_collections(name, dsk, dependencies=[self.obj])
+        return new_dd_object(graph, name, meta=meta, divisions=[iindexer, iindexer])
 
     def _get_partitions(self, keys):
         if isinstance(keys, (list, np.ndarray)):
@@ -179,7 +193,7 @@ class _LocIndexer(_IndexerBase):
         return _coerce_loc_index(self.obj.divisions, key)
 
     def _loc_slice(self, iindexer, cindexer):
-        name = 'loc-%s' % tokenize(iindexer, cindexer, self)
+        name = "loc-%s" % tokenize(iindexer, cindexer, self)
 
         assert isinstance(iindexer, slice)
         assert iindexer.step in (None, 1)
@@ -203,21 +217,41 @@ class _LocIndexer(_IndexerBase):
             istop = self._coerce_loc_index(iindexer.stop)
 
         if stop == start:
-            dsk = {(name, 0): (methods.loc, (self._name, start),
-                               slice(iindexer.start, iindexer.stop), cindexer)}
+            dsk = {
+                (name, 0): (
+                    methods.loc,
+                    (self._name, start),
+                    slice(iindexer.start, iindexer.stop),
+                    cindexer,
+                )
+            }
             divisions = [istart, istop]
         else:
-            dsk = {(name, 0): (methods.loc, (self._name, start),
-                               slice(iindexer.start, None), cindexer)}
+            dsk = {
+                (name, 0): (
+                    methods.loc,
+                    (self._name, start),
+                    slice(iindexer.start, None),
+                    cindexer,
+                )
+            }
             for i in range(1, stop - start):
                 if cindexer is None:
                     dsk[name, i] = (self._name, start + i)
                 else:
-                    dsk[name, i] = (methods.loc, (self._name, start + i),
-                                    slice(None, None), cindexer)
+                    dsk[name, i] = (
+                        methods.loc,
+                        (self._name, start + i),
+                        slice(None, None),
+                        cindexer,
+                    )
 
-            dsk[name, stop - start] = (methods.loc, (self._name, stop),
-                                       slice(None, iindexer.stop), cindexer)
+            dsk[name, stop - start] = (
+                methods.loc,
+                (self._name, stop),
+                slice(None, iindexer.stop),
+                cindexer,
+            )
 
             if iindexer.start is None:
                 div_start = self.obj.divisions[0]
@@ -229,15 +263,15 @@ class _LocIndexer(_IndexerBase):
             else:
                 div_stop = min(istop, self.obj.divisions[stop + 1])
 
-            divisions = ((div_start, ) +
-                         self.obj.divisions[start + 1:stop + 1] +
-                         (div_stop, ))
+            divisions = (
+                (div_start,) + self.obj.divisions[start + 1 : stop + 1] + (div_stop,)
+            )
 
         assert len(divisions) == len(dsk) + 1
 
         meta = self._make_meta(iindexer, cindexer)
-        return new_dd_object(merge(self.obj.dask, dsk), name,
-                             meta=meta, divisions=divisions)
+        graph = HighLevelGraph.from_collections(name, dsk, dependencies=[self.obj])
+        return new_dd_object(graph, name, meta=meta, divisions=divisions)
 
 
 def _partition_of_index_value(divisions, val):
@@ -301,26 +335,26 @@ def _maybe_partial_time_string(index, indexer, kind):
     if data has DatetimeIndex/PeriodIndex
     """
     # do not pass dd.Index
-    assert isinstance(index, pd.Index)
+    assert is_index_like(index)
 
     if not isinstance(index, (pd.DatetimeIndex, pd.PeriodIndex)):
         return indexer
 
     if isinstance(indexer, slice):
-        if isinstance(indexer.start, pd.compat.string_types):
-            start = index._maybe_cast_slice_bound(indexer.start, 'left', kind)
+        if isinstance(indexer.start, str):
+            start = index._maybe_cast_slice_bound(indexer.start, "left", kind)
         else:
             start = indexer.start
 
-        if isinstance(indexer.stop, pd.compat.string_types):
-            stop = index._maybe_cast_slice_bound(indexer.stop, 'right', kind)
+        if isinstance(indexer.stop, str):
+            stop = index._maybe_cast_slice_bound(indexer.stop, "right", kind)
         else:
             stop = indexer.stop
         return slice(start, stop)
 
-    elif isinstance(indexer, pd.compat.string_types):
-        start = index._maybe_cast_slice_bound(indexer, 'left', 'loc')
-        stop = index._maybe_cast_slice_bound(indexer, 'right', 'loc')
+    elif isinstance(indexer, str):
+        start = index._maybe_cast_slice_bound(indexer, "left", "loc")
+        stop = index._maybe_cast_slice_bound(indexer, "right", "loc")
         return slice(min(start, stop), max(start, stop))
 
     return indexer
