@@ -1,5 +1,4 @@
-from __future__ import print_function, division, absolute_import
-
+from collections.abc import Mapping
 from io import BytesIO
 from warnings import warn, catch_warnings, simplefilter
 
@@ -10,33 +9,36 @@ except ImportError:
 
 import numpy as np
 import pandas as pd
-try:
-    from pandas.api.types import CategoricalDtype
-    _HAS_CDT = True
-except ImportError:
-    _HAS_CDT = False
+from pandas.api.types import (
+    is_integer_dtype,
+    is_float_dtype,
+    is_object_dtype,
+    is_datetime64_any_dtype,
+    CategoricalDtype,
+)
 
-
-from ...bytes import read_bytes, open_files
-from ...bytes.compression import seekable_files, files as cfiles
-from ...compatibility import PY2, PY3, Mapping
+# this import checks for the importability of fsspec
+from ...bytes import read_bytes, open_file, open_files
 from ...delayed import delayed
-from ...utils import asciitable
-
-from ..utils import clear_known_categories, PANDAS_VERSION
-
+from ...utils import asciitable, parse_bytes
+from ..utils import clear_known_categories
 from .io import from_delayed
 
-if PANDAS_VERSION >= '0.20.0':
-    from pandas.api.types import (is_integer_dtype, is_float_dtype,
-                                  is_object_dtype, is_datetime64_any_dtype)
-else:
-    from pandas.types.common import (is_integer_dtype, is_float_dtype,
-                                     is_object_dtype, is_datetime64_any_dtype)
+import fsspec.implementations.local
+from fsspec.compression import compr
 
 
-def pandas_read_text(reader, b, header, kwargs, dtypes=None, columns=None,
-                     write_header=True, enforce=False, path=None):
+def pandas_read_text(
+    reader,
+    b,
+    header,
+    kwargs,
+    dtypes=None,
+    columns=None,
+    write_header=True,
+    enforce=False,
+    path=None,
+):
     """ Convert a block of bytes to a Pandas DataFrame
 
     Parameters
@@ -74,8 +76,9 @@ def pandas_read_text(reader, b, header, kwargs, dtypes=None, columns=None,
     if path:
         colname, path, paths = path
         code = paths.index(path)
-        df = df.assign(**{
-            colname: pd.Categorical.from_codes(np.full(len(df), code), paths)})
+        df = df.assign(
+            **{colname: pd.Categorical.from_codes(np.full(len(df), code), paths)}
+        )
     return df
 
 
@@ -113,23 +116,29 @@ def coerce_dtypes(df, dtypes):
 
     if bad_dtypes:
         if errors:
-            ex = '\n'.join("- %s\n  %r" % (c, e) for c, e in
-                           sorted(errors, key=lambda x: str(x[0])))
-            exceptions = ("The following columns also raised exceptions on "
-                          "conversion:\n\n%s\n\n") % ex
+            ex = "\n".join(
+                "- %s\n  %r" % (c, e)
+                for c, e in sorted(errors, key=lambda x: str(x[0]))
+            )
+            exceptions = (
+                "The following columns also raised exceptions on "
+                "conversion:\n\n%s\n\n"
+            ) % ex
             extra = ""
         else:
             exceptions = ""
             # All mismatches are int->float, also suggest `assume_missing=True`
-            extra = ("\n\nAlternatively, provide `assume_missing=True` "
-                     "to interpret\n"
-                     "all unspecified integer columns as floats.")
+            extra = (
+                "\n\nAlternatively, provide `assume_missing=True` "
+                "to interpret\n"
+                "all unspecified integer columns as floats."
+            )
 
         bad_dtypes = sorted(bad_dtypes, key=lambda x: str(x[0]))
-        table = asciitable(['Column', 'Found', 'Expected'], bad_dtypes)
-        dtype_kw = ('dtype={%s}' % ',\n'
-                    '       '.join("%r: '%s'" % (k, v)
-                                   for (k, v, _) in bad_dtypes))
+        table = asciitable(["Column", "Found", "Expected"], bad_dtypes)
+        dtype_kw = "dtype={%s}" % ",\n       ".join(
+            "%r: '%s'" % (k, v) for (k, v, _) in bad_dtypes
+        )
 
         dtype_msg = (
             "{table}\n\n"
@@ -138,34 +147,44 @@ def coerce_dtypes(df, dtypes):
             "*may* be fixed by specifying dtypes manually by adding:\n\n"
             "{dtype_kw}\n\n"
             "to the call to `read_csv`/`read_table`."
-            "{extra}").format(table=table, exceptions=exceptions,
-                              dtype_kw=dtype_kw, extra=extra)
+            "{extra}"
+        ).format(table=table, exceptions=exceptions, dtype_kw=dtype_kw, extra=extra)
     else:
         dtype_msg = None
 
     if bad_dates:
         also = " also " if bad_dtypes else " "
-        cols = '\n'.join("- %s" % c for c in bad_dates)
+        cols = "\n".join("- %s" % c for c in bad_dates)
         date_msg = (
             "The following columns{also}failed to properly parse as dates:\n\n"
             "{cols}\n\n"
             "This is usually due to an invalid value in that column. To\n"
             "diagnose and fix it's recommended to drop these columns from the\n"
             "`parse_dates` keyword, and manually convert them to dates later\n"
-            "using `dd.to_datetime`.").format(also=also, cols=cols)
+            "using `dd.to_datetime`."
+        ).format(also=also, cols=cols)
     else:
         date_msg = None
 
     if bad_dtypes or bad_dates:
-        rule = "\n\n%s\n\n" % ('-' * 61)
-        msg = ("Mismatched dtypes found in `pd.read_csv`/`pd.read_table`.\n\n"
-               "%s" % (rule.join(filter(None, [dtype_msg, date_msg]))))
+        rule = "\n\n%s\n\n" % ("-" * 61)
+        msg = "Mismatched dtypes found in `pd.read_csv`/`pd.read_table`.\n\n%s" % (
+            rule.join(filter(None, [dtype_msg, date_msg]))
+        )
         raise ValueError(msg)
 
 
-def text_blocks_to_pandas(reader, block_lists, header, head, kwargs,
-                          collection=True, enforce=False,
-                          specified_dtypes=None, path=None):
+def text_blocks_to_pandas(
+    reader,
+    block_lists,
+    header,
+    head,
+    kwargs,
+    collection=True,
+    enforce=False,
+    specified_dtypes=None,
+    path=None,
+):
     """ Convert blocks of bytes to a dask.dataframe or other high-level object
 
     This accepts a list of lists of values of bytes where each list corresponds
@@ -196,31 +215,32 @@ def text_blocks_to_pandas(reader, block_lists, header, head, kwargs,
     """
     dtypes = head.dtypes.to_dict()
     # dtypes contains only instances of CategoricalDtype, which causes issues
-    # in coerce_dtypes for non-uniform categories accross partitions.
+    # in coerce_dtypes for non-uniform categories across partitions.
     # We will modify `dtype` (which is inferred) to
     # 1. contain instances of CategoricalDtypes for user-provided types
     # 2. contain 'category' for data inferred types
-    categoricals = head.select_dtypes(include=['category']).columns
+    categoricals = head.select_dtypes(include=["category"]).columns
 
     known_categoricals = []
     unknown_categoricals = categoricals
-
-    if _HAS_CDT:
-        if isinstance(specified_dtypes, Mapping):
-            known_categoricals = [
-                k for k in categoricals
-                if isinstance(specified_dtypes.get(k), CategoricalDtype) and
-                specified_dtypes.get(k).categories is not None
-            ]
-            unknown_categoricals = categoricals.difference(known_categoricals)
-        elif isinstance(specified_dtypes, CategoricalDtype):
-            if specified_dtypes.categories is None:
-                known_categoricals = []
-                unknown_categoricals = categoricals
+    if isinstance(specified_dtypes, Mapping):
+        known_categoricals = [
+            k
+            for k in categoricals
+            if isinstance(specified_dtypes.get(k), CategoricalDtype)
+            and specified_dtypes.get(k).categories is not None
+        ]
+        unknown_categoricals = categoricals.difference(known_categoricals)
+    elif (
+        isinstance(specified_dtypes, CategoricalDtype)
+        and specified_dtypes.categories is None
+    ):
+        known_categoricals = []
+        unknown_categoricals = categoricals
 
     # Fixup the dtypes
     for k in unknown_categoricals:
-        dtypes[k] = 'category'
+        dtypes[k] = "category"
 
     columns = list(head.columns)
     delayed_pandas_read_text = delayed(pandas_read_text, pure=True)
@@ -234,23 +254,44 @@ def text_blocks_to_pandas(reader, block_lists, header, head, kwargs,
             path_info = (colname, paths[i], paths)
         else:
             path_info = None
-        df = delayed_pandas_read_text(reader, blocks[0], header, kwargs,
-                                      dtypes, columns, write_header=False,
-                                      enforce=enforce, path=path_info)
+        df = delayed_pandas_read_text(
+            reader,
+            blocks[0],
+            header,
+            kwargs,
+            dtypes,
+            columns,
+            write_header=False,
+            enforce=enforce,
+            path=path_info,
+        )
 
         dfs.append(df)
         rest_kwargs = kwargs.copy()
-        rest_kwargs.pop('skiprows', None)
+        rest_kwargs.pop("skiprows", None)
         for b in blocks[1:]:
-            dfs.append(delayed_pandas_read_text(reader, b, header, rest_kwargs,
-                                                dtypes, columns,
-                                                enforce=enforce, path=path_info))
+            dfs.append(
+                delayed_pandas_read_text(
+                    reader,
+                    b,
+                    header,
+                    rest_kwargs,
+                    dtypes,
+                    columns,
+                    enforce=enforce,
+                    path=path_info,
+                )
+            )
 
     if collection:
         if path:
-            head = head.assign(**{
-                colname: pd.Categorical.from_codes(np.zeros(len(head)), paths)
-            })
+            head = head.assign(
+                **{
+                    colname: pd.Categorical.from_codes(
+                        np.zeros(len(head), dtype=int), paths
+                    )
+                }
+            )
         if len(unknown_categoricals):
             head = clear_known_categories(head, cols=unknown_categoricals)
         return from_delayed(dfs, head)
@@ -272,73 +313,97 @@ if psutil is not None:
         CPU_COUNT = psutil.cpu_count()
         AUTO_BLOCKSIZE = auto_blocksize(TOTAL_MEM, CPU_COUNT)
 else:
-    AUTO_BLOCKSIZE = 2**25
+    AUTO_BLOCKSIZE = 2 ** 25
 
 
-def read_pandas(reader, urlpath, blocksize=AUTO_BLOCKSIZE, collection=True,
-                lineterminator=None, compression=None, sample=256000,
-                enforce=False, assume_missing=False, storage_options=None,
-                include_path_column=False,
-                **kwargs):
+def read_pandas(
+    reader,
+    urlpath,
+    blocksize=AUTO_BLOCKSIZE,
+    collection=True,
+    lineterminator=None,
+    compression=None,
+    sample=256000,
+    enforce=False,
+    assume_missing=False,
+    storage_options=None,
+    include_path_column=False,
+    **kwargs
+):
     reader_name = reader.__name__
     if lineterminator is not None and len(lineterminator) == 1:
-        kwargs['lineterminator'] = lineterminator
+        kwargs["lineterminator"] = lineterminator
     else:
-        lineterminator = '\n'
+        lineterminator = "\n"
     if include_path_column and isinstance(include_path_column, bool):
-        include_path_column = 'path'
-    if 'index' in kwargs or 'index_col' in kwargs:
-        raise ValueError("Keyword 'index' not supported "
-                         "dd.{0}(...).set_index('my-index') "
-                         "instead".format(reader_name))
-    for kw in ['iterator', 'chunksize']:
+        include_path_column = "path"
+    if "index" in kwargs or "index_col" in kwargs:
+        raise ValueError(
+            "Keywords 'index' and 'index_col' not supported. "
+            "Use dd.{0}(...).set_index('my-index') "
+            "instead".format(reader_name)
+        )
+    for kw in ["iterator", "chunksize"]:
         if kw in kwargs:
-            raise ValueError("{0} not supported for "
-                             "dd.{1}".format(kw, reader_name))
-    if kwargs.get('nrows', None):
-        raise ValueError("The 'nrows' keyword is not supported by "
-                         "`dd.{0}`. To achieve the same behavior, it's "
-                         "recommended to use `dd.{0}(...)."
-                         "head(n=nrows)`".format(reader_name))
-    if isinstance(kwargs.get('skiprows'), list):
+            raise ValueError("{0} not supported for dd.{1}".format(kw, reader_name))
+    if kwargs.get("nrows", None):
+        raise ValueError(
+            "The 'nrows' keyword is not supported by "
+            "`dd.{0}`. To achieve the same behavior, it's "
+            "recommended to use `dd.{0}(...)."
+            "head(n=nrows)`".format(reader_name)
+        )
+    if isinstance(kwargs.get("skiprows"), int):
+        skiprows = lastskiprow = firstrow = kwargs.get("skiprows")
+    elif kwargs.get("skiprows") is None:
+        skiprows = lastskiprow = firstrow = 0
+    else:
         # When skiprows is a list, we expect more than max(skiprows) to
         # be included in the sample. This means that [0,2] will work well,
         # but [0, 440] might not work.
-        skiprows = kwargs.get('skiprows')
+        skiprows = set(kwargs.get("skiprows"))
         lastskiprow = max(skiprows)
         # find the firstrow that is not skipped, for use as header
         firstrow = min(set(range(len(skiprows) + 1)) - set(skiprows))
-    else:
-        skiprows = lastskiprow = firstrow = kwargs.get('skiprows', 0)
-    if isinstance(kwargs.get('header'), list):
-        raise TypeError("List of header rows not supported for "
-                        "dd.{0}".format(reader_name))
-    if isinstance(kwargs.get('converters'), dict) and include_path_column:
-        path_converter = kwargs.get('converters').get(include_path_column, None)
+    if isinstance(kwargs.get("header"), list):
+        raise TypeError(
+            "List of header rows not supported for dd.{0}".format(reader_name)
+        )
+    if isinstance(kwargs.get("converters"), dict) and include_path_column:
+        path_converter = kwargs.get("converters").get(include_path_column, None)
     else:
         path_converter = None
 
-    if blocksize and compression not in seekable_files:
-        warn("Warning %s compression does not support breaking apart files\n"
-             "Please ensure that each individual file can fit in memory and\n"
-             "use the keyword ``blocksize=None to remove this message``\n"
-             "Setting ``blocksize=None``" % compression)
+    if isinstance(blocksize, str):
+        blocksize = parse_bytes(blocksize)
+    if blocksize and compression:
+        # NONE of the compressions should use chunking
+        warn(
+            "Warning %s compression does not support breaking apart files\n"
+            "Please ensure that each individual file can fit in memory and\n"
+            "use the keyword ``blocksize=None to remove this message``\n"
+            "Setting ``blocksize=None``" % compression
+        )
         blocksize = None
-    if compression not in seekable_files and compression not in cfiles:
-        raise NotImplementedError("Compression format %s not installed" %
-                                  compression)
-    if blocksize and blocksize < sample and lastskiprow != 0:
-        warn("Unexpected behavior can result from passing skiprows when\n"
-             "blocksize is smaller than sample size.\n"
-             "Setting ``sample=blocksize``")
+    if compression not in compr:
+        raise NotImplementedError("Compression format %s not installed" % compression)
+    if blocksize and sample and blocksize < sample and lastskiprow != 0:
+        warn(
+            "Unexpected behavior can result from passing skiprows when\n"
+            "blocksize is smaller than sample size.\n"
+            "Setting ``sample=blocksize``"
+        )
         sample = blocksize
     b_lineterminator = lineterminator.encode()
-    b_out = read_bytes(urlpath, delimiter=b_lineterminator,
-                       blocksize=blocksize,
-                       sample=sample,
-                       compression=compression,
-                       include_path=include_path_column,
-                       **(storage_options or {}))
+    b_out = read_bytes(
+        urlpath,
+        delimiter=b_lineterminator,
+        blocksize=blocksize,
+        sample=sample,
+        compression=compression,
+        include_path=include_path_column,
+        **(storage_options or {})
+    )
 
     if include_path_column:
         b_sample, values, paths = b_out
@@ -351,33 +416,40 @@ def read_pandas(reader, urlpath, blocksize=AUTO_BLOCKSIZE, collection=True,
 
     if not isinstance(values[0], (tuple, list)):
         values = [values]
+    # If we have not sampled, then use the first row of the first values
+    # as a representative sample.
+    if b_sample is False and len(values[0]):
+        b_sample = values[0][0].compute()
 
     # Get header row, and check that sample is long enough. If the file
     # contains a header row, we need at least 2 nonempty rows + the number of
     # rows to skip.
-    names = kwargs.get('names', None)
-    header = kwargs.get('header', 'infer' if names is None else None)
+    names = kwargs.get("names", None)
+    header = kwargs.get("header", "infer" if names is None else None)
     need = 1 if header is None else 2
     parts = b_sample.split(b_lineterminator, lastskiprow + need)
     # If the last partition is empty, don't count it
     nparts = 0 if not parts else len(parts) - int(not parts[-1])
 
-    if nparts < lastskiprow + need and len(b_sample) >= sample:
-        raise ValueError("Sample is not large enough to include at least one "
-                         "row of data. Please increase the number of bytes "
-                         "in `sample` in the call to `read_csv`/`read_table`")
+    if sample is not False and nparts < lastskiprow + need and len(b_sample) >= sample:
+        raise ValueError(
+            "Sample is not large enough to include at least one "
+            "row of data. Please increase the number of bytes "
+            "in `sample` in the call to `read_csv`/`read_table`"
+        )
 
-    header = b'' if header is None else parts[firstrow] + b_lineterminator
+    header = b"" if header is None else parts[firstrow] + b_lineterminator
 
-    # Use sample to infer dtypes and check for presense of include_path_column
+    # Use sample to infer dtypes and check for presence of include_path_column
     head = reader(BytesIO(b_sample), **kwargs)
     if include_path_column and (include_path_column in head.columns):
-        raise ValueError("Files already contain the column name: %s, so the "
-                         "path column cannot use this name. Please set "
-                         "`include_path_column` to a unique name."
-                         % include_path_column)
+        raise ValueError(
+            "Files already contain the column name: %s, so the "
+            "path column cannot use this name. Please set "
+            "`include_path_column` to a unique name." % include_path_column
+        )
 
-    specified_dtypes = kwargs.get('dtype', {})
+    specified_dtypes = kwargs.get("dtype", {})
     if specified_dtypes is None:
         specified_dtypes = {}
     # If specified_dtypes is a single type, then all columns were specified
@@ -387,10 +459,17 @@ def read_pandas(reader, urlpath, blocksize=AUTO_BLOCKSIZE, collection=True,
             if is_integer_dtype(head[c].dtype) and c not in specified_dtypes:
                 head[c] = head[c].astype(float)
 
-    return text_blocks_to_pandas(reader, values, header, head, kwargs,
-                                 collection=collection, enforce=enforce,
-                                 specified_dtypes=specified_dtypes,
-                                 path=path)
+    return text_blocks_to_pandas(
+        reader,
+        values,
+        header,
+        head,
+        kwargs,
+        collection=collection,
+        enforce=enforce,
+        specified_dtypes=specified_dtypes,
+        path=path,
+    )
 
 
 READ_DOC_TEMPLATE = """
@@ -424,10 +503,11 @@ urlpath : string or list
     to read from alternative filesystems. To read from multiple files you
     can pass a globstring or a list of paths, with the caveat that they
     must all have the same protocol.
-blocksize : int or None, optional
+blocksize : str, int or None, optional
     Number of bytes by which to cut up larger files. Default value is
     computed based on available physical memory and the number of cores.
     If ``None``, use a single block for each file.
+    Can be a number like 64000000 or a string like "64MB"
 collection : boolean, optional
     Return a dask.dataframe if True or list of dask.delayed objects if False
 sample : int, optional
@@ -470,37 +550,64 @@ at the cost of reduced parallelism.
 
 
 def make_reader(reader, reader_name, file_type):
-    def read(urlpath, blocksize=AUTO_BLOCKSIZE, collection=True,
-             lineterminator=None, compression=None, sample=256000,
-             enforce=False, assume_missing=False, storage_options=None,
-             include_path_column=False,
-             **kwargs):
-        return read_pandas(reader, urlpath, blocksize=blocksize,
-                           collection=collection,
-                           lineterminator=lineterminator,
-                           compression=compression, sample=sample,
-                           enforce=enforce, assume_missing=assume_missing,
-                           storage_options=storage_options,
-                           include_path_column=include_path_column,
-                           **kwargs)
-    read.__doc__ = READ_DOC_TEMPLATE.format(reader=reader_name,
-                                            file_type=file_type)
+    def read(
+        urlpath,
+        blocksize=AUTO_BLOCKSIZE,
+        collection=True,
+        lineterminator=None,
+        compression=None,
+        sample=256000,
+        enforce=False,
+        assume_missing=False,
+        storage_options=None,
+        include_path_column=False,
+        **kwargs
+    ):
+        return read_pandas(
+            reader,
+            urlpath,
+            blocksize=blocksize,
+            collection=collection,
+            lineterminator=lineterminator,
+            compression=compression,
+            sample=sample,
+            enforce=enforce,
+            assume_missing=assume_missing,
+            storage_options=storage_options,
+            include_path_column=include_path_column,
+            **kwargs
+        )
+
+    read.__doc__ = READ_DOC_TEMPLATE.format(reader=reader_name, file_type=file_type)
     read.__name__ = reader_name
     return read
 
 
-read_csv = make_reader(pd.read_csv, 'read_csv', 'CSV')
-read_table = make_reader(pd.read_table, 'read_table', 'delimited')
+read_csv = make_reader(pd.read_csv, "read_csv", "CSV")
+read_table = make_reader(pd.read_table, "read_table", "delimited")
+read_fwf = make_reader(pd.read_fwf, "read_fwf", "fixed-width")
 
 
-def _write_csv(df, fil, **kwargs):
+def _write_csv(df, fil, *, depend_on=None, **kwargs):
     with fil as f:
         df.to_csv(f, **kwargs)
+    return None
 
 
-def to_csv(df, filename, name_function=None, compression=None, compute=True,
-           scheduler=None, storage_options=None,
-           header_first_partition_only=False, **kwargs):
+def to_csv(
+    df,
+    filename,
+    single_file=False,
+    encoding="utf-8",
+    mode="wt",
+    name_function=None,
+    compression=None,
+    compute=True,
+    scheduler=None,
+    storage_options=None,
+    header_first_partition_only=None,
+    **kwargs
+):
     """
     Store Dask DataFrame to CSV files
 
@@ -552,7 +659,15 @@ def to_csv(df, filename, name_function=None, compression=None, compute=True,
     name_function : callable, default None
         Function accepting an integer (partition index) and producing a
         string to replace the asterisk in the given filename globstring.
-        Should preserve the lexicographic order of partitions
+        Should preserve the lexicographic order of partitions. Not
+        supported when `single_file` is `True`.
+    single_file : bool, default False
+        Whether to save everything into a single CSV file. Under the
+        single file mode, each partition is appended at the end of the
+        specified CSV file. Note that not all filesystems support the
+        append mode and thus the single file mode, especially on cloud
+        storage systems such as S3 or GCS. A warning will be issued when
+        writing to a file that is not backed by a local filesystem.
     compression : string or None
         String like 'gzip' or 'xz'.  Must support efficient random access.
         Filenames with extensions corresponding to known compression
@@ -568,8 +683,12 @@ def to_csv(df, filename, name_function=None, compression=None, compute=True,
     header : boolean or list of string, default True
         Write out column names. If a list of string is given it is assumed
         to be aliases for the column names
-    header_first_partition_only : boolean, default False
-        If set, only write the header row in the first output file
+    header_first_partition_only : boolean, default None
+        If set to `True`, only write the header row in the first output
+        file. By default, headers are written to all partitions under
+        the multiple file mode (`single_file` is `False`) and written
+        only once under the single file mode (`single_file` is `True`).
+        It must not be `False` under the single file mode.
     index : boolean, default True
         Write row names (index)
     index_label : string or sequence, or False, default None
@@ -617,35 +736,55 @@ def to_csv(df, filename, name_function=None, compression=None, compute=True,
     -------
     The names of the file written if they were computed right away
     If not, the delayed tasks associated to the writing of the files
+
+    Raises
+    ------
+    ValueError
+        If `header_first_partition_only` is set to `False` or
+        `name_function` is specified when `single_file` is `True`.
     """
-    if PY2:
-        default_encoding = None
-        mode = 'wb'
-    else:
-        default_encoding = 'utf-8'
-        mode = 'wt'
-
-    encoding = kwargs.get('encoding', default_encoding)
-
-    files = open_files(filename, compression=compression, mode=mode,
-                       encoding=encoding, name_function=name_function,
-                       num=df.npartitions, **(storage_options or {}))
-
+    if single_file and name_function is not None:
+        raise ValueError("name_function is not supported under the single file mode")
+    if header_first_partition_only is None:
+        header_first_partition_only = single_file
+    elif not header_first_partition_only and single_file:
+        raise ValueError(
+            "header_first_partition_only cannot be False in the single file mode."
+        )
+    file_options = dict(
+        compression=compression,
+        encoding=encoding,
+        newline="",
+        **(storage_options or {})
+    )
     to_csv_chunk = delayed(_write_csv, pure=False)
-
     dfs = df.to_delayed()
-
-    # If we only want headers in the first file, turn headers off after the
-    # first partition
-
-    values = [to_csv_chunk(dfs[0], files[0], **kwargs)]
-
-    if len(dfs) > 1:
+    if single_file:
+        first_file = open_file(filename, mode=mode, **file_options)
+        if not isinstance(first_file.fs, fsspec.implementations.local.LocalFileSystem):
+            warn("Appending data to a network storage system may not work.")
+        value = to_csv_chunk(dfs[0], first_file, **kwargs)
+        append_mode = mode.replace("w", "") + "a"
+        append_file = open_file(filename, mode=append_mode, **file_options)
+        kwargs["header"] = False
+        for d in dfs[1:]:
+            value = to_csv_chunk(d, append_file, depend_on=value, **kwargs)
+        values = [value]
+        files = [first_file]
+    else:
+        files = open_files(
+            filename,
+            mode=mode,
+            name_function=name_function,
+            num=df.npartitions,
+            **file_options
+        )
+        values = [to_csv_chunk(dfs[0], files[0], **kwargs)]
         if header_first_partition_only:
-            kwargs['header'] = False
-        values.extend([to_csv_chunk(d, f, **kwargs)
-                       for d, f in zip(dfs[1:], files[1:])])
-
+            kwargs["header"] = False
+        values.extend(
+            [to_csv_chunk(d, f, **kwargs) for d, f in zip(dfs[1:], files[1:])]
+        )
     if compute:
         delayed(values).compute(scheduler=scheduler)
         return [f.path for f in files]
@@ -653,6 +792,6 @@ def to_csv(df, filename, name_function=None, compression=None, compute=True,
         return values
 
 
-if PY3:
-    from ..core import _Frame
-    _Frame.to_csv.__doc__ = to_csv.__doc__
+from ..core import _Frame
+
+_Frame.to_csv.__doc__ = to_csv.__doc__
