@@ -1,5 +1,3 @@
-from __future__ import absolute_import, division, print_function
-
 from math import ceil
 from operator import getitem
 import os
@@ -10,11 +8,10 @@ import numpy as np
 from toolz import merge
 
 from ...base import tokenize
-from ...compatibility import unicode, PY3
 from ... import array as da
 from ...delayed import delayed
 
-from ..core import DataFrame, Series, Index, new_dd_object
+from ..core import DataFrame, Series, Index, new_dd_object, has_parallel_type
 from ..shuffle import set_partition
 from ..utils import insert_meta_param_description, check_meta, make_meta
 
@@ -23,14 +20,21 @@ from ...utils import M, ensure_dict
 lock = Lock()
 
 
-def _meta_from_array(x, columns=None):
+def _meta_from_array(x, columns=None, index=None):
     """ Create empty pd.DataFrame or pd.Series which has correct dtype """
 
     if x.ndim > 2:
-        raise ValueError('from_array does not input more than 2D array, got'
-                         ' array with shape %r' % (x.shape,))
+        raise ValueError(
+            "from_array does not input more than 2D array, got"
+            " array with shape %r" % (x.shape,)
+        )
 
-    if getattr(x.dtype, 'names', None) is not None:
+    if index is not None:
+        if not isinstance(index, Index):
+            raise ValueError("'index' must be an instance of dask.dataframe.Index")
+        index = index._meta
+
+    if getattr(x.dtype, "names", None) is not None:
         # record array has named columns
         if columns is None:
             columns = list(x.dtype.names)
@@ -38,30 +42,34 @@ def _meta_from_array(x, columns=None):
             raise ValueError("For a struct dtype, columns must be a list.")
         elif not all(i in x.dtype.names for i in columns):
             extra = sorted(set(columns).difference(x.dtype.names))
-            raise ValueError("dtype {0} doesn't have fields "
-                             "{1}".format(x.dtype, extra))
+            raise ValueError("dtype {0} doesn't have fields {1}".format(x.dtype, extra))
         fields = x.dtype.fields
-        dtypes = [fields[n][0] if n in fields else 'f8' for n in columns]
+        dtypes = [fields[n][0] if n in fields else "f8" for n in columns]
     elif x.ndim == 1:
         if np.isscalar(columns) or columns is None:
-            return pd.Series([], name=columns, dtype=x.dtype)
+            return pd.Series([], name=columns, dtype=x.dtype, index=index)
         elif len(columns) == 1:
-            return pd.DataFrame(np.array([], dtype=x.dtype), columns=columns)
-        raise ValueError("For a 1d array, columns must be a scalar or single "
-                         "element list")
+            return pd.DataFrame(
+                np.array([], dtype=x.dtype), columns=columns, index=index
+            )
+        raise ValueError(
+            "For a 1d array, columns must be a scalar or single element list"
+        )
     else:
         if np.isnan(x.shape[1]):
             raise ValueError("Shape along axis 1 must be known")
         if columns is None:
             columns = list(range(x.shape[1])) if x.ndim == 2 else [0]
         elif len(columns) != x.shape[1]:
-            raise ValueError("Number of column names must match width of the "
-                             "array. Got {0} names for {1} "
-                             "columns".format(len(columns), x.shape[1]))
+            raise ValueError(
+                "Number of column names must match width of the "
+                "array. Got {0} names for {1} "
+                "columns".format(len(columns), x.shape[1])
+            )
         dtypes = [x.dtype] * len(columns)
 
     data = {c: np.array([], dtype=dt) for (c, dt) in zip(columns, dtypes)}
-    return pd.DataFrame(data, columns=columns)
+    return pd.DataFrame(data, columns=columns, index=index)
 
 
 def from_array(x, chunksize=50000, columns=None):
@@ -89,7 +97,7 @@ def from_array(x, chunksize=50000, columns=None):
     divisions = tuple(range(0, len(x), chunksize))
     divisions = divisions + (len(x) - 1,)
     token = tokenize(x, chunksize, columns)
-    name = 'from_array-' + token
+    name = "from_array-" + token
 
     dsk = {}
     for i in range(0, int(ceil(len(x) / chunksize))):
@@ -115,7 +123,7 @@ def from_pandas(data, npartitions=None, chunksize=None, sort=True, name=None):
 
     Parameters
     ----------
-    df : pandas.DataFrame or pandas.Series
+    data : pandas.DataFrame or pandas.Series
         The DataFrame/Series with which to construct a Dask DataFrame/Series
     npartitions : int, optional
         The number of partitions of the index to create. Note that depending on
@@ -162,14 +170,14 @@ def from_pandas(data, npartitions=None, chunksize=None, sort=True, name=None):
     from_array : Construct a dask.DataFrame from an array that has record dtype
     read_csv : Construct a dask.DataFrame from a CSV file
     """
-    if isinstance(getattr(data, 'index', None), pd.MultiIndex):
+    if isinstance(getattr(data, "index", None), pd.MultiIndex):
         raise NotImplementedError("Dask does not support MultiIndex Dataframes.")
 
-    if not isinstance(data, (pd.Series, pd.DataFrame)):
+    if not has_parallel_type(data):
         raise TypeError("Input must be a pandas DataFrame or Series")
 
-    if ((npartitions is None) == (chunksize is None)):
-        raise ValueError('Exactly one of npartitions and chunksize must be specified.')
+    if (npartitions is None) == (chunksize is None):
+        raise ValueError("Exactly one of npartitions and chunksize must be specified.")
 
     nrows = len(data)
 
@@ -178,7 +186,7 @@ def from_pandas(data, npartitions=None, chunksize=None, sort=True, name=None):
     else:
         npartitions = int(ceil(nrows / chunksize))
 
-    name = name or ('from_pandas-' + tokenize(data, chunksize))
+    name = name or ("from_pandas-" + tokenize(data, chunksize))
 
     if not nrows:
         return new_dd_object({(name, 0): data}, name, data, [None, None])
@@ -186,20 +194,21 @@ def from_pandas(data, npartitions=None, chunksize=None, sort=True, name=None):
     if sort and not data.index.is_monotonic_increasing:
         data = data.sort_index(ascending=True)
     if sort:
-        divisions, locations = sorted_division_locations(data.index,
-                                                         chunksize=chunksize)
+        divisions, locations = sorted_division_locations(
+            data.index, chunksize=chunksize
+        )
     else:
         locations = list(range(0, nrows, chunksize)) + [len(data)]
         divisions = [None] * len(locations)
 
-    dsk = dict(((name, i), data.iloc[start: stop])
-               for i, (start, stop) in enumerate(zip(locations[:-1],
-                                                     locations[1:])))
+    dsk = {
+        (name, i): data.iloc[start:stop]
+        for i, (start, stop) in enumerate(zip(locations[:-1], locations[1:]))
+    }
     return new_dd_object(dsk, name, data, divisions)
 
 
-def from_bcolz(x, chunksize=None, categorize=True, index=None, lock=lock,
-               **kwargs):
+def from_bcolz(x, chunksize=None, categorize=True, index=None, lock=lock, **kwargs):
     """ Read BColz CTable into a Dask Dataframe
 
     BColz is a fast on-disk compressed column store with careful attention
@@ -227,7 +236,7 @@ def from_bcolz(x, chunksize=None, categorize=True, index=None, lock=lock,
     import dask.array as da
     import bcolz
 
-    if isinstance(x, (str, unicode)):
+    if isinstance(x, str):
         x = bcolz.ctable(rootdir=x)
     bc_chunklen = max(x[name].chunklen for name in x.names)
     if chunksize is None and bc_chunklen > 10000:
@@ -236,9 +245,11 @@ def from_bcolz(x, chunksize=None, categorize=True, index=None, lock=lock,
     categories = dict()
     if categorize:
         for name in x.names:
-            if (np.issubdtype(x.dtype[name], np.string_) or
-                    np.issubdtype(x.dtype[name], np.unicode_) or
-                    np.issubdtype(x.dtype[name], np.object_)):
+            if (
+                np.issubdtype(x.dtype[name], np.string_)
+                or np.issubdtype(x.dtype[name], np.unicode_)
+                or np.issubdtype(x.dtype[name], np.object_)
+            ):
                 a = da.from_array(x[name], chunks=(chunksize * len(x.names),))
                 categories[name] = da.unique(a).compute()
 
@@ -246,19 +257,33 @@ def from_bcolz(x, chunksize=None, categorize=True, index=None, lock=lock,
     divisions = tuple(range(0, len(x), chunksize))
     divisions = divisions + (len(x) - 1,)
     if x.rootdir:
-        token = tokenize((x.rootdir, os.path.getmtime(x.rootdir)), chunksize,
-                         categorize, index, kwargs)
+        token = tokenize(
+            (x.rootdir, os.path.getmtime(x.rootdir)),
+            chunksize,
+            categorize,
+            index,
+            kwargs,
+        )
     else:
-        token = tokenize((id(x), x.shape, x.dtype), chunksize, categorize,
-                         index, kwargs)
-    new_name = 'from_bcolz-' + token
+        token = tokenize(
+            (id(x), x.shape, x.dtype), chunksize, categorize, index, kwargs
+        )
+    new_name = "from_bcolz-" + token
 
-    dsk = dict(((new_name, i),
-                (dataframe_from_ctable,
-                 x,
-                 (slice(i * chunksize, (i + 1) * chunksize),),
-                 columns, categories, lock))
-               for i in range(0, int(ceil(len(x) / chunksize))))
+    dsk = dict(
+        (
+            (new_name, i),
+            (
+                dataframe_from_ctable,
+                x,
+                (slice(i * chunksize, (i + 1) * chunksize),),
+                columns,
+                categories,
+                lock,
+            ),
+        )
+        for i in range(0, int(ceil(len(x) / chunksize)))
+    )
 
     meta = dataframe_from_ctable(x, slice(0, 0), columns, categories, lock)
     result = DataFrame(dsk, new_name, meta, divisions)
@@ -301,6 +326,7 @@ def dataframe_from_ctable(x, slc, columns=None, categories=None, lock=lock):
 
     """
     import bcolz
+
     if columns is None:
         columns = x.dtype.names
     if isinstance(columns, tuple):
@@ -321,20 +347,26 @@ def dataframe_from_ctable(x, slc, columns=None, categories=None, lock=lock):
         if isinstance(x, bcolz.ctable):
             chunks = [x[name][slc] for name in columns]
             if categories is not None:
-                chunks = [pd.Categorical.from_codes(
-                    np.searchsorted(categories[name], chunk),
-                    categories[name], True)
-                    if name in categories else chunk
-                    for name, chunk in zip(columns, chunks)]
-            result = pd.DataFrame(dict(zip(columns, chunks)), columns=columns,
-                                  index=idx)
+                chunks = [
+                    pd.Categorical.from_codes(
+                        np.searchsorted(categories[name], chunk), categories[name], True
+                    )
+                    if name in categories
+                    else chunk
+                    for name, chunk in zip(columns, chunks)
+                ]
+            result = pd.DataFrame(
+                dict(zip(columns, chunks)), columns=columns, index=idx
+            )
 
         elif isinstance(x, bcolz.carray):
             chunk = x[slc]
             if categories is not None and columns and columns in categories:
                 chunk = pd.Categorical.from_codes(
                     np.searchsorted(categories[columns], chunk),
-                    categories[columns], True)
+                    categories[columns],
+                    True,
+                )
             result = pd.Series(chunk, name=columns, index=idx)
     finally:
         if lock:
@@ -383,12 +415,12 @@ def from_dask_array(x, columns=None, index=None):
     dask.dataframe._Frame.values: Reverse conversion
     dask.dataframe._Frame.to_records: Reverse conversion
     """
-    meta = _meta_from_array(x, columns)
+    meta = _meta_from_array(x, columns, index)
 
     if x.ndim == 2 and len(x.chunks[1]) > 1:
         x = x.rechunk({1: x.shape[1]})
 
-    name = 'from-dask-array' + tokenize(x, columns)
+    name = "from-dask-array" + tokenize(x, columns)
     to_merge = []
 
     if index is not None:
@@ -396,7 +428,7 @@ def from_dask_array(x, columns=None, index=None):
             raise ValueError("'index' must be an instance of dask.dataframe.Index")
         if index.npartitions != x.numblocks[0]:
             msg = (
-                "'index' must have the same number of blocks as 'x'. "
+                "The index and array have different numbers of blocks. "
                 "({} != {})".format(index.npartitions, x.numblocks[0])
             )
             raise ValueError(msg)
@@ -411,8 +443,9 @@ def from_dask_array(x, columns=None, index=None):
         divisions = [0]
         for c in x.chunks[0]:
             divisions.append(divisions[-1] + c)
-        index = [(np.arange, a, b, 1, 'i8') for a, b in
-                 zip(divisions[:-1], divisions[1:])]
+        index = [
+            (np.arange, a, b, 1, "i8") for a, b in zip(divisions[:-1], divisions[1:])
+        ]
         divisions[-1] -= 1
 
     dsk = {}
@@ -458,11 +491,14 @@ def to_bag(df, index=False):
     >>> bag = df.to_bag()  # doctest: +SKIP
     """
     from ...bag.core import Bag
+
     if not isinstance(df, (DataFrame, Series)):
         raise TypeError("df must be either DataFrame or Series")
-    name = 'to_bag-' + tokenize(df, index)
-    dsk = dict(((name, i), (_df_to_bag, block, index))
-               for (i, block) in enumerate(df.__dask_keys__()))
+    name = "to_bag-" + tokenize(df, index)
+    dsk = dict(
+        ((name, i), (_df_to_bag, block, index))
+        for (i, block) in enumerate(df.__dask_keys__())
+    )
     dsk.update(df.__dask_optimize__(df.__dask_graph__(), df.__dask_keys__()))
     return Bag(dsk, name, df.npartitions)
 
@@ -477,7 +513,7 @@ def to_records(df):
     Examples
     --------
     >>> df.to_records()  # doctest: +SKIP
-    dask.array<shape=(nan,), dtype=(numpy.record, [('ind', '<f8'), ('x', 'O'), ('y', '<i8')]), chunksize=(nan,)>
+    dask.array<to_records, shape=(nan,), dtype=(numpy.record, [('ind', '<f8'), ('x', 'O'), ('y', '<i8')]), chunksize=(nan,), chunktype=numpy.ndarray>  # noqa: E501
 
     See Also
     --------
@@ -488,7 +524,9 @@ def to_records(df):
 
 
 @insert_meta_param_description
-def from_delayed(dfs, meta=None, divisions=None, prefix='from-delayed'):
+def from_delayed(
+    dfs, meta=None, divisions=None, prefix="from-delayed", verify_meta=True
+):
     """ Create Dask DataFrame from many Dask Delayed objects
 
     Parameters
@@ -506,29 +544,39 @@ def from_delayed(dfs, meta=None, divisions=None, prefix='from-delayed'):
         If None, then won't use index information
     prefix : str, optional
         Prefix to prepend to the keys.
+    verify_meta : bool, optional
+        If True check that the partitions have consistent metadata, defaults to True.
     """
     from dask.delayed import Delayed
+
     if isinstance(dfs, Delayed):
         dfs = [dfs]
-    dfs = [delayed(df)
-           if not isinstance(df, Delayed) and hasattr(df, 'key')
-           else df
-           for df in dfs]
+    dfs = [
+        delayed(df) if not isinstance(df, Delayed) and hasattr(df, "key") else df
+        for df in dfs
+    ]
     for df in dfs:
         if not isinstance(df, Delayed):
-            raise TypeError("Expected Delayed object, got %s" %
-                            type(df).__name__)
+            raise TypeError("Expected Delayed object, got %s" % type(df).__name__)
 
     if meta is None:
-        meta = dfs[0].compute()
-    meta = make_meta(meta)
+        meta = delayed(make_meta)(dfs[0]).compute()
+    else:
+        meta = make_meta(meta)
 
-    name = prefix + '-' + tokenize(*dfs)
+    name = prefix + "-" + tokenize(*dfs)
     dsk = merge(df.dask for df in dfs)
-    dsk.update({(name, i): (check_meta, df.key, meta, 'from_delayed')
-                for (i, df) in enumerate(dfs)})
+    if verify_meta:
+        dsk.update(
+            {
+                (name, i): (check_meta, df.key, meta, "from_delayed")
+                for (i, df) in enumerate(dfs)
+            }
+        )
+    else:
+        dsk.update({(name, i): df.key for (i, df) in enumerate(dfs)})
 
-    if divisions is None or divisions == 'sorted':
+    if divisions is None or divisions == "sorted":
         divs = [None] * (len(dfs) + 1)
     else:
         divs = tuple(divisions)
@@ -537,8 +585,9 @@ def from_delayed(dfs, meta=None, divisions=None, prefix='from-delayed'):
 
     df = new_dd_object(dsk, name, meta, divs)
 
-    if divisions == 'sorted':
+    if divisions == "sorted":
         from ..shuffle import compute_divisions
+
         divisions = compute_divisions(df)
         df.divisions = divisions
 
@@ -568,8 +617,8 @@ def sorted_division_locations(seq, npartitions=None, chunksize=None):
     >>> sorted_division_locations(['A'], chunksize=2)
     (['A', 'A'], [0, 1])
     """
-    if ((npartitions is None) == (chunksize is None)):
-        raise ValueError('Exactly one of npartitions and chunksize must be specified.')
+    if (npartitions is None) == (chunksize is None):
+        raise ValueError("Exactly one of npartitions and chunksize must be specified.")
 
     if npartitions:
         chunksize = ceil(len(seq) / npartitions)
@@ -593,6 +642,5 @@ def sorted_division_locations(seq, npartitions=None, chunksize=None):
     return values, positions
 
 
-if PY3:
-    DataFrame.to_records.__doc__ = to_records.__doc__
-    DataFrame.to_bag.__doc__ = to_bag.__doc__
+DataFrame.to_records.__doc__ = to_records.__doc__
+DataFrame.to_bag.__doc__ = to_bag.__doc__
