@@ -1,13 +1,17 @@
-from __future__ import absolute_import, division, print_function
-
 import math
+import numbers
 import re
-from operator import getitem
 
-from . import config
-from .compatibility import unicode
-from .core import (istask, get_dependencies, subs, toposort,
-                   flatten, reverse_dict, ishashable, _get_recursive)
+from . import config, core
+from .core import (
+    istask,
+    get_dependencies,
+    subs,
+    toposort,
+    flatten,
+    reverse_dict,
+    ishashable,
+)
 from .utils_test import add, inc  # noqa: F401
 
 
@@ -42,8 +46,10 @@ def cull(dsk, keys):
     while work:
         new_work = []
         out_keys += work
-        deps = [(k, get_dependencies(dsk, k, as_list=True))  # fuse needs lists
-                for k in work]
+        deps = [
+            (k, get_dependencies(dsk, k, as_list=True))  # fuse needs lists
+            for k in work
+        ]
         dependencies.update(deps)
         for _, deplist in deps:
             for d in deplist:
@@ -60,15 +66,14 @@ def cull(dsk, keys):
 def default_fused_linear_keys_renamer(keys):
     """Create new keys for fused tasks"""
     typ = type(keys[0])
-    if typ is str or typ is unicode:
+    if typ is str:
         names = [key_split(x) for x in keys[:0:-1]]
         names.append(keys[0])
-        return '-'.join(names)
-    elif (typ is tuple and len(keys[0]) > 0 and
-          isinstance(keys[0][0], (str, unicode))):
+        return "-".join(names)
+    elif typ is tuple and len(keys[0]) > 0 and isinstance(keys[0][0], str):
         names = [key_split(x) for x in keys[:0:-1]]
         names.append(keys[0][0])
-        return ('-'.join(names),) + keys[0][1:]
+        return ("-".join(names),) + keys[0][1:]
     else:
         return None
 
@@ -121,8 +126,7 @@ def fuse_linear(dsk, keys=None, dependencies=None, rename_keys=True):
         keys = set(flatten(keys))
 
     if dependencies is None:
-        dependencies = {k: get_dependencies(dsk, k, as_list=True)
-                        for k in dsk}
+        dependencies = {k: get_dependencies(dsk, k, as_list=True) for k in dsk}
 
     # locate all members of linear chains
     child2parent = {}
@@ -175,8 +179,9 @@ def fuse_linear(dsk, keys=None, dependencies=None, rename_keys=True):
     for chain in chains:
         if key_renamer is not None:
             new_key = key_renamer(chain)
-            is_renamed = (new_key is not None and new_key not in dsk and
-                          new_key not in rv)
+            is_renamed = (
+                new_key is not None and new_key not in dsk and new_key not in rv
+            )
         child = chain.pop()
         val = dsk[child]
         while chain:
@@ -247,18 +252,20 @@ def inline(dsk, keys=None, inline_constants=True, dependencies=None):
     keys = _flat_set(keys)
 
     if dependencies is None:
-        dependencies = {k: get_dependencies(dsk, k)
-                        for k in dsk}
+        dependencies = {k: get_dependencies(dsk, k) for k in dsk}
 
     if inline_constants:
-        keys.update(k for k, v in dsk.items() if
-                    (ishashable(v) and v in dsk) or
-                    (not dependencies[k] and not istask(v)))
+        keys.update(
+            k
+            for k, v in dsk.items()
+            if (ishashable(v) and v in dsk) or (not dependencies[k] and not istask(v))
+        )
 
     # Keys may depend on other keys, so determine replace order with toposort.
     # The values stored in `keysubs` do not include other keys.
-    replaceorder = toposort(dict((k, dsk[k]) for k in keys if k in dsk),
-                            dependencies=dependencies)
+    replaceorder = toposort(
+        dict((k, dsk[k]) for k in keys if k in dsk), dependencies=dependencies
+    )
     keysubs = {}
     for key in replaceorder:
         val = dsk[key]
@@ -280,8 +287,9 @@ def inline(dsk, keys=None, inline_constants=True, dependencies=None):
     return dsk2
 
 
-def inline_functions(dsk, output, fast_functions=None, inline_constants=False,
-                     dependencies=None):
+def inline_functions(
+    dsk, output, fast_functions=None, inline_constants=False, dependencies=None
+):
     """ Inline cheap functions into larger operations
 
     Examples
@@ -311,8 +319,7 @@ def inline_functions(dsk, output, fast_functions=None, inline_constants=False,
     fast_functions = set(fast_functions)
 
     if dependencies is None:
-        dependencies = {k: get_dependencies(dsk, k)
-                        for k in dsk}
+        dependencies = {k: get_dependencies(dsk, k) for k in dsk}
     dependents = reverse_dict(dependencies)
 
     def inlinable(v):
@@ -321,19 +328,23 @@ def inline_functions(dsk, output, fast_functions=None, inline_constants=False,
         except TypeError:
             return False
 
-    keys = [k for k, v in dsk.items()
-            if istask(v) and dependents[k] and k not in output and inlinable(v)]
+    keys = [
+        k
+        for k, v in dsk.items()
+        if istask(v) and dependents[k] and k not in output and inlinable(v)
+    ]
 
     if keys:
-        dsk = inline(dsk, keys, inline_constants=inline_constants,
-                     dependencies=dependencies)
+        dsk = inline(
+            dsk, keys, inline_constants=inline_constants, dependencies=dependencies
+        )
         for k in keys:
             del dsk[k]
     return dsk
 
 
 def unwrap_partial(func):
-    while hasattr(func, 'func'):
+    while hasattr(func, "func"):
         func = func.func
     return func
 
@@ -366,99 +377,38 @@ def functions_of(task):
     return funcs
 
 
-def fuse_selections(dsk, head1, head2, merge):
-    """Fuse selections with lower operation.
-
-    Handles graphs of the form:
-    ``{key1: (head1, key2, ...), key2: (head2, ...)}``
-
-    Parameters
-    ----------
-    dsk : dict
-        dask graph
-    head1 : function
-        The first element of task1
-    head2 : function
-        The first element of task2
-    merge : function
-        Takes ``task1`` and ``task2`` and returns a merged task to
-        replace ``task1``.
-
-    Examples
-    --------
-    >>> def load(store, partition, columns):
-    ...     pass
-    >>> dsk = {'x': (load, 'store', 'part', ['a', 'b']),
-    ...        'y': (getitem, 'x', 'a')}
-    >>> merge = lambda t1, t2: (load, t2[1], t2[2], t1[2])
-    >>> dsk2 = fuse_selections(dsk, getitem, load, merge)
-    >>> cull(dsk2, 'y')[0]
-    {'y': (<function load at ...>, 'store', 'part', 'a')}
-    """
-    dsk2 = dict()
-    for k, v in dsk.items():
-        try:
-            if (istask(v) and v[0] == head1 and v[1] in dsk and
-                    istask(dsk[v[1]]) and dsk[v[1]][0] == head2):
-                dsk2[k] = merge(v, dsk[v[1]])
-            else:
-                dsk2[k] = v
-        except TypeError:
-            dsk2[k] = v
-    return dsk2
-
-
-def fuse_getitem(dsk, func, place):
-    """ Fuse getitem with lower operation
-
-    Parameters
-    ----------
-    dsk: dict
-        dask graph
-    func: function
-        A function in a task to merge
-    place: int
-        Location in task to insert the getitem key
-
-    Examples
-    --------
-    >>> def load(store, partition, columns):
-    ...     pass
-    >>> dsk = {'x': (load, 'store', 'part', ['a', 'b']),
-    ...        'y': (getitem, 'x', 'a')}
-    >>> dsk2 = fuse_getitem(dsk, load, 3)  # columns in arg place 3
-    >>> cull(dsk2, 'y')[0]
-    {'y': (<function load at ...>, 'store', 'part', 'a')}
-    """
-    return fuse_selections(dsk, getitem, func,
-                           lambda a, b: tuple(b[:place]) + (a[2], ) + tuple(b[place + 1:]))
-
-
 def default_fused_keys_renamer(keys):
     """Create new keys for ``fuse`` tasks"""
     it = reversed(keys)
     first_key = next(it)
     typ = type(first_key)
-    if typ is str or typ is unicode:
+    if typ is str:
         first_name = key_split(first_key)
         names = {key_split(k) for k in it}
         names.discard(first_name)
         names = sorted(names)
         names.append(first_key)
-        return '-'.join(names)
-    elif (typ is tuple and len(first_key) > 0 and
-          isinstance(first_key[0], (str, unicode))):
+        return "-".join(names)
+    elif typ is tuple and len(first_key) > 0 and isinstance(first_key[0], str):
         first_name = key_split(first_key)
         names = {key_split(k) for k in it}
         names.discard(first_name)
         names = sorted(names)
         names.append(first_key[0])
-        return ('-'.join(names),) + first_key[1:]
+        return ("-".join(names),) + first_key[1:]
 
 
-def fuse(dsk, keys=None, dependencies=None, ave_width=None, max_width=None,
-         max_height=None, max_depth_new_edges=None, rename_keys=None,
-         fuse_subgraphs=None):
+def fuse(
+    dsk,
+    keys=None,
+    dependencies=None,
+    ave_width=None,
+    max_width=None,
+    max_height=None,
+    max_depth_new_edges=None,
+    rename_keys=None,
+    fuse_subgraphs=None,
+):
     """ Fuse tasks that form reductions; more advanced than ``fuse_linear``
 
     This trades parallelism opportunities for faster scheduling by making tasks
@@ -470,7 +420,7 @@ def fuse(dsk, keys=None, dependencies=None, ave_width=None, max_width=None,
     tune the behavior, which are described below.  ``ave_width`` is the
     natural parameter with which to compare parallelism to granularity, so
     it should always be specified.  Reasonable values for other parameters
-    with be determined using ``ave_width`` if necessary.
+    will be determined using ``ave_width`` if necessary.
 
     Parameters
     ----------
@@ -512,36 +462,36 @@ def fuse(dsk, keys=None, dependencies=None, ave_width=None, max_width=None,
 
     # Assign reasonable, not too restrictive defaults
     if ave_width is None:
-        if config.get('fuse_ave_width', None) is None:
+        if config.get("fuse_ave_width", None) is None:
             ave_width = 1
         else:
-            ave_width = config.get('fuse_ave_width', None)
+            ave_width = config.get("fuse_ave_width", None)
 
     if max_height is None:
-        if config.get('fuse_max_height', None) is None:
+        if config.get("fuse_max_height", None) is None:
             max_height = len(dsk)
         else:
-            max_height = config.get('fuse_max_height', None)
+            max_height = config.get("fuse_max_height", None)
 
     max_depth_new_edges = (
-        max_depth_new_edges or
-        config.get('fuse_max_depth_new_edges', None) or
-        ave_width + 1.5
+        max_depth_new_edges
+        or config.get("fuse_max_depth_new_edges", None)
+        or ave_width + 1.5
     )
     max_width = (
-        max_width or
-        config.get('fuse_max_width', None) or
-        1.5 + ave_width * math.log(ave_width + 1)
+        max_width
+        or config.get("fuse_max_width", None)
+        or 1.5 + ave_width * math.log(ave_width + 1)
     )
 
     if fuse_subgraphs is None:
-        fuse_subgraphs = config.get('fuse_subgraphs', False)
+        fuse_subgraphs = config.get("fuse_subgraphs", False)
 
     if not ave_width or not max_height:
         return dsk, dependencies
 
     if rename_keys is None:
-        rename_keys = config.get('fuse_rename_keys', True)
+        rename_keys = config.get("fuse_rename_keys", True)
     if rename_keys is True:
         key_renamer = default_fused_keys_renamer
     elif rename_keys is False:
@@ -567,8 +517,14 @@ def fuse(dsk, keys=None, dependencies=None, ave_width=None, max_width=None,
     reducible = {k for k, vals in rdeps.items() if len(vals) == 1}
     if keys:
         reducible -= keys
-    if (not reducible and
-            (not fuse_subgraphs or all(len(set(v)) != 1 for v in rdeps.values()))):
+
+    for k, v in dsk.items():
+        if type(v) is not tuple and not isinstance(v, (numbers.Number, str)):
+            reducible.discard(k)
+
+    if not reducible and (
+        not fuse_subgraphs or all(len(set(v)) != 1 for v in rdeps.values())
+    ):
         # Quick return if there's nothing to do. Only progress if there's tasks
         # fusible by the main `fuse`, or by `fuse_subgraphs` if enabled.
         return dsk, deps
@@ -610,8 +566,18 @@ def fuse(dsk, keys=None, dependencies=None, ave_width=None, max_width=None,
                 children_stack_pop()
                 # This is a leaf node in the reduction region
                 # key, task, fused_keys, height, width, number of nodes, fudge, set of edges
-                info_stack_append((child, rv[child], [child] if rename_keys else None,
-                                   1, 1, 1, 0, deps[child] - reducible))
+                info_stack_append(
+                    (
+                        child,
+                        rv[child],
+                        [child] if rename_keys else None,
+                        1,
+                        1,
+                        1,
+                        0,
+                        deps[child] - reducible,
+                    )
+                )
             else:
                 children_stack_pop()
                 # Calculate metrics and fuse as appropriate
@@ -621,8 +587,16 @@ def fuse(dsk, keys=None, dependencies=None, ave_width=None, max_width=None,
                 num_children = len(children)
 
                 if num_children == 1:
-                    (child_key, child_task, child_keys, height, width, num_nodes, fudge,
-                     children_edges) = info_stack_pop()
+                    (
+                        child_key,
+                        child_task,
+                        child_keys,
+                        height,
+                        width,
+                        num_nodes,
+                        fudge,
+                        children_edges,
+                    ) = info_stack_pop()
                     num_children_edges = len(children_edges)
 
                     if fudge > num_children_edges - 1 >= 0:
@@ -632,7 +606,8 @@ def fuse(dsk, keys=None, dependencies=None, ave_width=None, max_width=None,
                     if not no_new_edges:
                         fudge += 1
                     if (
-                        (num_nodes + fudge) / height <= ave_width and
+                        (num_nodes + fudge) / height <= ave_width
+                        and
                         # Sanity check; don't go too deep if new levels introduce new edge dependencies
                         (no_new_edges or height < max_depth_new_edges)
                     ):
@@ -650,10 +625,31 @@ def fuse(dsk, keys=None, dependencies=None, ave_width=None, max_width=None,
                         if children_stack:
                             if no_new_edges:
                                 # Linear fuse
-                                info_stack_append((parent, val, child_keys, height, width, num_nodes, fudge, edges))
+                                info_stack_append(
+                                    (
+                                        parent,
+                                        val,
+                                        child_keys,
+                                        height,
+                                        width,
+                                        num_nodes,
+                                        fudge,
+                                        edges,
+                                    )
+                                )
                             else:
-                                info_stack_append((parent, val, child_keys, height + 1, width, num_nodes + 1, fudge,
-                                                   edges))
+                                info_stack_append(
+                                    (
+                                        parent,
+                                        val,
+                                        child_keys,
+                                        height + 1,
+                                        width,
+                                        num_nodes + 1,
+                                        fudge,
+                                        edges,
+                                    )
+                                )
                         else:
                             rv[parent] = val
                             break
@@ -666,8 +662,18 @@ def fuse(dsk, keys=None, dependencies=None, ave_width=None, max_width=None,
                             if fudge > int(ave_width - 1):
                                 fudge = int(ave_width - 1)
                             # This task *implicitly* depends on `edges`
-                            info_stack_append((parent, rv[parent], [parent] if rename_keys else None,
-                                               1, width, 1, fudge, edges))
+                            info_stack_append(
+                                (
+                                    parent,
+                                    rv[parent],
+                                    [parent] if rename_keys else None,
+                                    1,
+                                    width,
+                                    1,
+                                    fudge,
+                                    edges,
+                                )
+                            )
                         else:
                             break
                 else:
@@ -681,8 +687,16 @@ def fuse(dsk, keys=None, dependencies=None, ave_width=None, max_width=None,
                     max_num_edges = 0
                     children_info = info_stack[-num_children:]
                     del info_stack[-num_children:]
-                    for cur_key, cur_task, cur_keys, cur_height, cur_width, cur_num_nodes, cur_fudge, \
-                            cur_edges in children_info:
+                    for (
+                        cur_key,
+                        cur_task,
+                        cur_keys,
+                        cur_height,
+                        cur_width,
+                        cur_num_nodes,
+                        cur_fudge,
+                        cur_edges,
+                    ) in children_info:
                         if cur_height == 1:
                             num_single_nodes += 1
                         elif cur_height > height:
@@ -695,7 +709,9 @@ def fuse(dsk, keys=None, dependencies=None, ave_width=None, max_width=None,
                         children_edges |= cur_edges
                     # Fudge factor to account for possible parallelism with the boundaries
                     num_children_edges = len(children_edges)
-                    fudge += min(num_children - 1, max(0, num_children_edges - max_num_edges))
+                    fudge += min(
+                        num_children - 1, max(0, num_children_edges - max_num_edges)
+                    )
 
                     if fudge > num_children_edges - 1 >= 0:
                         fudge = num_children_edges - 1
@@ -704,10 +720,11 @@ def fuse(dsk, keys=None, dependencies=None, ave_width=None, max_width=None,
                     if not no_new_edges:
                         fudge += 1
                     if (
-                        (num_nodes + fudge) / height <= ave_width and
-                        num_single_nodes <= ave_width and
-                        width <= max_width and
-                        height <= max_height and
+                        (num_nodes + fudge) / height <= ave_width
+                        and num_single_nodes <= ave_width
+                        and width <= max_width
+                        and height <= max_height
+                        and
                         # Sanity check; don't go too deep if new levels introduce new edge dependencies
                         (no_new_edges or height < max_depth_new_edges)
                     ):
@@ -731,7 +748,18 @@ def fuse(dsk, keys=None, dependencies=None, ave_width=None, max_width=None,
                             fused_trees[parent] = child_keys
 
                         if children_stack:
-                            info_stack_append((parent, val, child_keys, height + 1, width, num_nodes + 1, fudge, edges))
+                            info_stack_append(
+                                (
+                                    parent,
+                                    val,
+                                    child_keys,
+                                    height + 1,
+                                    width,
+                                    num_nodes + 1,
+                                    fudge,
+                                    edges,
+                                )
+                            )
                         else:
                             rv[parent] = val
                             break
@@ -748,8 +776,18 @@ def fuse(dsk, keys=None, dependencies=None, ave_width=None, max_width=None,
                                 fudge = int(ave_width - 1)
                             # key, task, height, width, number of nodes, fudge, set of edges
                             # This task *implicitly* depends on `edges`
-                            info_stack_append((parent, rv[parent], [parent] if rename_keys else None,
-                                               1, width, 1, fudge, edges))
+                            info_stack_append(
+                                (
+                                    parent,
+                                    rv[parent],
+                                    [parent] if rename_keys else None,
+                                    1,
+                                    width,
+                                    1,
+                                    fudge,
+                                    edges,
+                                )
+                            )
                         else:
                             break
                 # Traverse upwards
@@ -843,7 +881,7 @@ def _inplace_fuse_subgraphs(dsk, keys, dependencies, fused_trees, rename_keys):
 
 # Defining `key_split` (used by key renamers in `fuse`) in utils.py
 # results in messy circular imports, so define it here instead.
-hex_pattern = re.compile('[a-f]+')
+hex_pattern = re.compile("[a-f]+")
 
 
 def key_split(s):
@@ -878,25 +916,26 @@ def key_split(s):
     if type(s) is tuple:
         s = s[0]
     try:
-        words = s.split('-')
+        words = s.split("-")
         if not words[0][0].isalpha():
             result = words[0].strip("_'()\"")
         else:
             result = words[0]
         for word in words[1:]:
-            if word.isalpha() and not (len(word) == 8 and
-                                       hex_pattern.match(word) is not None):
-                result += '-' + word
+            if word.isalpha() and not (
+                len(word) == 8 and hex_pattern.match(word) is not None
+            ):
+                result += "-" + word
             else:
                 break
-        if len(result) == 32 and re.match(r'[a-f0-9]{32}', result):
-            return 'data'
+        if len(result) == 32 and re.match(r"[a-f0-9]{32}", result):
+            return "data"
         else:
-            if result[0] == '<':
-                result = result.strip('<>').split()[0].split('.')[-1]
+            if result[0] == "<":
+                result = result.strip("<>").split()[0].split(".")[-1]
             return result
     except Exception:
-        return 'Other'
+        return "Other"
 
 
 class SubgraphCallable(object):
@@ -913,9 +952,10 @@ class SubgraphCallable(object):
     name : str, optional
         The name to use for the function.
     """
-    __slots__ = ('dsk', 'outkey', 'inkeys', 'name')
 
-    def __init__(self, dsk, outkey, inkeys, name='subgraph_callable'):
+    __slots__ = ("dsk", "outkey", "inkeys", "name")
+
+    def __init__(self, dsk, outkey, inkeys, name="subgraph_callable"):
         self.dsk = dsk
         self.outkey = outkey
         self.inkeys = inkeys
@@ -925,22 +965,21 @@ class SubgraphCallable(object):
         return self.name
 
     def __eq__(self, other):
-        return (type(self) is type(other) and
-                self.dsk == other.dsk and
-                self.outkey == other.outkey and
-                set(self.inkeys) == set(other.inkeys) and
-                self.name == other.name)
+        return (
+            type(self) is type(other)
+            and self.dsk == other.dsk
+            and self.outkey == other.outkey
+            and set(self.inkeys) == set(other.inkeys)
+            and self.name == other.name
+        )
 
     def __ne__(self, other):
         return not (self == other)
 
     def __call__(self, *args):
         if not len(args) == len(self.inkeys):
-            raise ValueError("Expected %d args, got %d"
-                             % (len(self.inkeys), len(args)))
-        return _get_recursive(self.dsk, self.outkey,
-                              dict(zip(self.inkeys, args)))
+            raise ValueError("Expected %d args, got %d" % (len(self.inkeys), len(args)))
+        return core.get(self.dsk, self.outkey, dict(zip(self.inkeys, args)))
 
     def __reduce__(self):
-        return (SubgraphCallable,
-                (self.dsk, self.outkey, self.inkeys, self.name))
+        return (SubgraphCallable, (self.dsk, self.outkey, self.inkeys, self.name))
