@@ -1,5 +1,4 @@
-from __future__ import absolute_import, division, print_function
-
+from datetime import timedelta
 import functools
 import inspect
 import os
@@ -8,22 +7,28 @@ import sys
 import tempfile
 import re
 from errno import ENOENT
+from collections.abc import Iterator
 from contextlib import contextmanager
 from importlib import import_module
-from numbers import Integral
+from numbers import Integral, Number
 from threading import Lock
 import uuid
 from weakref import WeakValueDictionary
 
-from .compatibility import (get_named_args, getargspec, PY3, unicode,
-                            bind_method, Iterator)
 from .core import get_deps
-from .optimization import key_split    # noqa: F401
+from .optimization import key_split  # noqa: F401
 
 
 system_encoding = sys.getdefaultencoding()
-if system_encoding == 'ascii':
-    system_encoding = 'utf-8'
+if system_encoding == "ascii":
+    system_encoding = "utf-8"
+
+
+def apply(func, args, kwargs=None):
+    if kwargs:
+        return func(*args, **kwargs)
+    else:
+        return func(*args)
 
 
 def deepmap(func, *seqs):
@@ -94,8 +99,8 @@ def import_required(mod_name, error_msg):
 
 
 @contextmanager
-def tmpfile(extension='', dir=None):
-    extension = '.' + extension.lstrip('.')
+def tmpfile(extension="", dir=None):
+    extension = "." + extension.lstrip(".")
     handle, filename = tempfile.mkstemp(extension, dir=dir)
     os.close(handle)
     os.remove(filename)
@@ -128,7 +133,7 @@ def tmpdir(dir=None):
 
 
 @contextmanager
-def filetext(text, extension='', open=open, mode='w'):
+def filetext(text, extension="", open=open, mode="w"):
     with tmpfile(extension=extension) as filename:
         f = open(filename, mode=mode)
         try:
@@ -174,7 +179,8 @@ class IndexCallable(object):
     >>> I[3]
     4
     """
-    __slots__ = 'fn',
+
+    __slots__ = ("fn",)
 
     def __init__(self, fn):
         self.fn = fn
@@ -184,7 +190,7 @@ class IndexCallable(object):
 
 
 @contextmanager
-def filetexts(d, open=open, mode='t', use_tmpdir=True):
+def filetexts(d, open=open, mode="t", use_tmpdir=True):
     """ Dumps a number of textfiles to disk
 
     d - dict
@@ -196,7 +202,11 @@ def filetexts(d, open=open, mode='t', use_tmpdir=True):
     """
     with (tmp_cwd() if use_tmpdir else noop_context()):
         for filename, text in d.items():
-            f = open(filename, 'w' + mode)
+            try:
+                os.makedirs(os.path.dirname(filename))
+            except OSError:
+                pass
+            f = open(filename, "w" + mode)
             try:
                 f.write(text)
             finally:
@@ -238,6 +248,7 @@ def pseudorandom(n, p, random_state=None):
     array([0, 2, 0, 3, 0, 1, 2, 1, 0, 0], dtype=int8)
     """
     import numpy as np
+
     p = list(p)
     cp = np.cumsum([0] + p)
     assert np.allclose(1, cp[-1])
@@ -247,7 +258,7 @@ def pseudorandom(n, p, random_state=None):
         random_state = np.random.RandomState(random_state)
 
     x = random_state.random_sample(n)
-    out = np.empty(n, dtype='i1')
+    out = np.empty(n, dtype="i1")
 
     for i, (low, high) in enumerate(zip(cp[:-1], cp[1:])):
         out[(x >= low) & (x < high)] = i
@@ -267,7 +278,9 @@ def random_state_data(n, random_state=None):
     """
     import numpy as np
 
-    if not all(hasattr(random_state, attr) for attr in ['normal', 'beta', 'bytes', 'uniform']):
+    if not all(
+        hasattr(random_state, attr) for attr in ["normal", "beta", "bytes", "uniform"]
+    ):
         random_state = np.random.RandomState(random_state)
 
     random_data = random_state.bytes(624 * n * 4)  # `n * 624` 32-bit integers
@@ -288,17 +301,83 @@ def is_integer(i):
     return isinstance(i, Integral) or (isinstance(i, float) and i.is_integer())
 
 
-ONE_ARITY_BUILTINS = set([abs, all, any, bool, bytearray, bytes, callable, chr,
-                          classmethod, complex, dict, dir, enumerate, eval,
-                          float, format, frozenset, hash, hex, id, int, iter,
-                          len, list, max, min, next, oct, open, ord, range,
-                          repr, reversed, round, set, slice, sorted,
-                          staticmethod, str, sum, tuple,
-                          type, vars, zip, memoryview])
-if PY3:
-    ONE_ARITY_BUILTINS.add(ascii)  # noqa: F821
-MULTI_ARITY_BUILTINS = set([compile, delattr, divmod, filter, getattr, hasattr,
-                            isinstance, issubclass, map, pow, setattr])
+ONE_ARITY_BUILTINS = set(
+    [
+        abs,
+        all,
+        any,
+        ascii,
+        bool,
+        bytearray,
+        bytes,
+        callable,
+        chr,
+        classmethod,
+        complex,
+        dict,
+        dir,
+        enumerate,
+        eval,
+        float,
+        format,
+        frozenset,
+        hash,
+        hex,
+        id,
+        int,
+        iter,
+        len,
+        list,
+        max,
+        min,
+        next,
+        oct,
+        open,
+        ord,
+        range,
+        repr,
+        reversed,
+        round,
+        set,
+        slice,
+        sorted,
+        staticmethod,
+        str,
+        sum,
+        tuple,
+        type,
+        vars,
+        zip,
+        memoryview,
+    ]
+)
+MULTI_ARITY_BUILTINS = set(
+    [
+        compile,
+        delattr,
+        divmod,
+        filter,
+        getattr,
+        hasattr,
+        isinstance,
+        issubclass,
+        map,
+        pow,
+        setattr,
+    ]
+)
+
+
+def getargspec(func):
+    """Version of inspect.getargspec that works with partial and warps."""
+    if isinstance(func, functools.partial):
+        return getargspec(func.func)
+
+    func = getattr(func, "__wrapped__", func)
+    if isinstance(func, type):
+        return inspect.getfullargspec(func.__init__)
+    else:
+        return inspect.getfullargspec(func)
 
 
 def takes_multiple_arguments(func, varargs=True):
@@ -337,7 +416,7 @@ def takes_multiple_arguments(func, varargs=True):
         return False
 
     try:
-        is_constructor = spec.args[0] == 'self' and isinstance(func, type)
+        is_constructor = spec.args[0] == "self" and isinstance(func, type)
     except Exception:
         is_constructor = False
 
@@ -348,8 +427,19 @@ def takes_multiple_arguments(func, varargs=True):
     return len(spec.args) - ndefaults - is_constructor > 1
 
 
+def get_named_args(func):
+    """Get all non ``*args/**kwargs`` arguments for a function"""
+    s = inspect.signature(func)
+    return [
+        n
+        for n, p in s.parameters.items()
+        if p.kind in [p.POSITIONAL_OR_KEYWORD, p.POSITIONAL_ONLY, p.KEYWORD_ONLY]
+    ]
+
+
 class Dispatch(object):
     """Simple single dispatch."""
+
     def __init__(self, name=None):
         self._lookup = {}
         self._lazy = {}
@@ -358,6 +448,7 @@ class Dispatch(object):
 
     def register(self, type, func=None):
         """Register dispatch of `func` on arguments of type `type`"""
+
         def wrapper(func):
             if isinstance(type, tuple):
                 for t in type:
@@ -373,6 +464,7 @@ class Dispatch(object):
         Register a registration function which will be called if the
         *toplevel* module (e.g. 'pandas') is ever loaded.
         """
+
         def wrapper(func):
             self._lazy[toplevel] = func
             return func
@@ -390,14 +482,14 @@ class Dispatch(object):
         else:
             return impl
         # Is a lazy registration function present?
-        toplevel, _, _ = cls.__module__.partition('.')
+        toplevel, _, _ = cls.__module__.partition(".")
         try:
             register = self._lazy.pop(toplevel)
         except KeyError:
             pass
         else:
             register()
-            return self.dispatch(cls) # recurse
+            return self.dispatch(cls)  # recurse
         # Walk the MRO and cache the lookup result
         for cls2 in inspect.getmro(cls)[1:]:
             if cls2 in lk:
@@ -435,41 +527,136 @@ def ensure_not_exists(filename):
 def _skip_doctest(line):
     # NumPy docstring contains cursor and comment only example
     stripped = line.strip()
-    if stripped == '>>>' or stripped.startswith('>>> #'):
-        return stripped
-    elif '>>>' in stripped and '+SKIP' not in stripped:
-        return line + '  # doctest: +SKIP'
+    if stripped == ">>>" or stripped.startswith(">>> #"):
+        return line
+    elif ">>>" in stripped and "+SKIP" not in stripped:
+        if "# doctest:" in line:
+            return line + ", +SKIP"
+        else:
+            return line + "  # doctest: +SKIP"
     else:
         return line
 
 
 def skip_doctest(doc):
     if doc is None:
-        return ''
-    return '\n'.join([_skip_doctest(line) for line in doc.split('\n')])
+        return ""
+    return "\n".join([_skip_doctest(line) for line in doc.split("\n")])
 
 
 def extra_titles(doc):
-    lines = doc.split('\n')
-    titles = {i: lines[i].strip() for i in range(len(lines) - 1)
-              if lines[i + 1] and all(c == '-' for c in lines[i + 1].strip())}
+    lines = doc.split("\n")
+    titles = {
+        i: lines[i].strip()
+        for i in range(len(lines) - 1)
+        if lines[i + 1].strip() and all(c == "-" for c in lines[i + 1].strip())
+    }
 
     seen = set()
     for i, title in sorted(titles.items()):
         if title in seen:
-            new_title = 'Extra ' + title
+            new_title = "Extra " + title
             lines[i] = lines[i].replace(title, new_title)
-            lines[i + 1] = lines[i + 1].replace('-' * len(title),
-                                                '-' * len(new_title))
+            lines[i + 1] = lines[i + 1].replace("-" * len(title), "-" * len(new_title))
         else:
             seen.add(title)
 
-    return '\n'.join(lines)
+    return "\n".join(lines)
+
+
+def ignore_warning(doc, cls, name, extra=""):
+    """Expand docstring by adding disclaimer and extra text"""
+    import inspect
+
+    if inspect.isclass(cls):
+        l1 = "This docstring was copied from %s.%s.%s. \n\n" % (
+            cls.__module__,
+            cls.__name__,
+            name,
+        )
+    else:
+        l1 = "This docstring was copied from %s.%s. \n\n" % (cls.__name__, name)
+    l2 = "Some inconsistencies with the Dask version may exist."
+
+    i = doc.find("\n\n")
+    if i != -1:
+        # Insert our warning
+        head = doc[: i + 2]
+        tail = doc[i + 2 :]
+        # Indentation of next line
+        indent = re.match(r"\s*", tail).group(0)
+        # Insert the warning, indented, with a blank line before and after
+        if extra:
+            more = [indent, extra.rstrip("\n") + "\n\n"]
+        else:
+            more = []
+        bits = [head, indent, l1, indent, l2, "\n\n"] + more + [tail]
+        doc = "".join(bits)
+
+    return doc
+
+
+def unsupported_arguments(doc, args):
+    """ Mark unsupported arguments with a disclaimer """
+    lines = doc.split("\n")
+    for arg in args:
+        subset = [
+            (i, line)
+            for i, line in enumerate(lines)
+            if re.match(r"^\s*" + arg + " ?:", line)
+        ]
+        if len(subset) == 1:
+            [(i, line)] = subset
+            lines[i] = line + "  (Not supported in Dask)"
+    return "\n".join(lines)
+
+
+def _derived_from(cls, method, ua_args=[], extra=""):
+    """ Helper function for derived_from to ease testing """
+    # do not use wraps here, as it hides keyword arguments displayed
+    # in the doc
+    original_method = getattr(cls, method.__name__)
+
+    if isinstance(original_method, property):
+        # some things like SeriesGroupBy.unique are generated.
+        original_method = original_method.fget
+
+    doc = original_method.__doc__
+    if doc is None:
+        doc = ""
+
+    # Insert disclaimer that this is a copied docstring
+    if doc:
+        doc = ignore_warning(doc, cls, method.__name__, extra=extra)
+    elif extra:
+        doc += extra.rstrip("\n") + "\n\n"
+
+    # Mark unsupported arguments
+    try:
+        method_args = get_named_args(method)
+        original_args = get_named_args(original_method)
+        not_supported = [m for m in original_args if m not in method_args]
+    except ValueError:
+        not_supported = []
+    if len(ua_args) > 0:
+        not_supported.extend(ua_args)
+    if len(not_supported) > 0:
+        doc = unsupported_arguments(doc, not_supported)
+
+    doc = skip_doctest(doc)
+    doc = extra_titles(doc)
+
+    return doc
 
 
 def derived_from(original_klass, version=None, ua_args=[]):
     """Decorator to attach original class's docstring to the wrapped method.
 
+    The output structure will be: top line of docstring, disclaimer about this
+    being auto-derived, any extra text associated with the method being patched,
+    the body of the docstring and finally, the list of keywords that exist in
+    the original method but not in the dask version.
+
     Parameters
     ----------
     original_klass: type
@@ -480,49 +667,28 @@ def derived_from(original_klass, version=None, ua_args=[]):
         List of keywords which Dask doesn't support. Keywords existing in
         original but not in Dask will automatically be added.
     """
+
     def wrapper(method):
-        method_name = method.__name__
-
         try:
-            # do not use wraps here, as it hides keyword arguments displayed
-            # in the doc
-            original_method = getattr(original_klass, method_name)
-            doc = original_method.__doc__
-            if doc is None:
-                doc = ''
-
-            try:
-                method_args = get_named_args(method)
-                original_args = get_named_args(original_method)
-                not_supported = [m for m in original_args if m not in method_args]
-            except ValueError:
-                not_supported = []
-
-            if len(ua_args) > 0:
-                not_supported.extend(ua_args)
-
-            if len(not_supported) > 0:
-                note = ("\n        Notes\n        -----\n"
-                        "        Dask doesn't support the following argument(s).\n\n")
-                args = ''.join(['        * {0}\n'.format(a) for a in not_supported])
-                doc = doc + note + args
-            doc = skip_doctest(doc)
-            doc = extra_titles(doc)
-
-            method.__doc__ = doc
+            extra = getattr(method, "__doc__", None) or ""
+            method.__doc__ = _derived_from(
+                original_klass, method, ua_args=ua_args, extra=extra
+            )
             return method
 
         except AttributeError:
-            module_name = original_klass.__module__.split('.')[0]
+            module_name = original_klass.__module__.split(".")[0]
 
             @functools.wraps(method)
             def wrapped(*args, **kwargs):
-                msg = "Base package doesn't support '{0}'.".format(method_name)
+                msg = "Base package doesn't support '{0}'.".format(method.__name__)
                 if version is not None:
                     msg2 = " Use {0} {1} or later to use this method."
                     msg += msg2.format(module_name, version)
                 raise NotImplementedError(msg)
+
             return wrapped
+
     return wrapper
 
 
@@ -533,26 +699,48 @@ def funcname(func):
         return funcname(func.func)
     # methodcaller
     if isinstance(func, methodcaller):
-        return func.method
+        return func.method[:50]
 
-    module_name = getattr(func, '__module__', None) or ''
-    type_name = getattr(type(func), '__name__', None) or ''
+    module_name = getattr(func, "__module__", None) or ""
+    type_name = getattr(type(func), "__name__", None) or ""
 
     # toolz.curry
-    if 'toolz' in module_name and 'curry' == type_name:
-        return func.func_name
+    if "toolz" in module_name and "curry" == type_name:
+        return func.func_name[:50]
     # multipledispatch objects
-    if 'multipledispatch' in module_name and 'Dispatcher' == type_name:
-        return func.name
+    if "multipledispatch" in module_name and "Dispatcher" == type_name:
+        return func.name[:50]
+    # numpy.vectorize objects
+    if "numpy" in module_name and "vectorize" == type_name:
+        return ("vectorize_" + funcname(func.pyfunc))[:50]
 
     # All other callables
     try:
         name = func.__name__
-        if name == '<lambda>':
-            return 'lambda'
-        return name
+        if name == "<lambda>":
+            return "lambda"
+        return name[:50]
     except AttributeError:
-        return str(func)
+        return str(func)[:50]
+
+
+def typename(typ):
+    """
+    Return the name of a type
+
+    Examples
+    --------
+    >>> typename(int)
+    'int'
+
+    >>> from dask.core import literal
+    >>> typename(literal)
+    'dask.core.literal'
+    """
+    if not typ.__module__ or typ.__module__ == "builtins":
+        return typ.__name__
+    else:
+        return typ.__module__ + "." + typ.__name__
 
 
 def ensure_bytes(s):
@@ -567,7 +755,7 @@ def ensure_bytes(s):
     """
     if isinstance(s, bytes):
         return s
-    if hasattr(s, 'encode'):
+    if hasattr(s, "encode"):
         return s.encode()
     msg = "Object %s is neither a bytes object nor has an encode method"
     raise TypeError(msg % s)
@@ -583,9 +771,9 @@ def ensure_unicode(s):
     >>> ensure_unicode(b'123')
     '123'
     """
-    if isinstance(s, unicode):
+    if isinstance(s, str):
         return s
-    if hasattr(s, 'decode'):
+    if hasattr(s, "decode"):
         return s.decode()
     msg = "Object %s is neither a bytes object nor has an encode method"
     raise TypeError(msg % s)
@@ -603,7 +791,7 @@ def digit(n, k, base):
     >>> digit(1234, 3, 10)
     1
     """
-    return n // base**k % base
+    return n // base ** k % base
 
 
 def insert(tup, loc, val):
@@ -634,7 +822,7 @@ def dependency_depth(dsk):
 
 
 def memory_repr(num):
-    for x in ['bytes', 'KB', 'MB', 'GB', 'TB']:
+    for x in ["bytes", "KB", "MB", "GB", "TB"]:
         if num < 1024.0:
             return "%3.1f %s" % (num, x)
         num /= 1024.0
@@ -653,19 +841,18 @@ def asciitable(columns, rows):
     """
     rows = [tuple(str(i) for i in r) for r in rows]
     columns = tuple(str(i) for i in columns)
-    widths = tuple(max(max(map(len, x)), len(c))
-                   for x, c in zip(zip(*rows), columns))
-    row_template = ('|' + (' %%-%ds |' * len(columns))) % widths
+    widths = tuple(max(max(map(len, x)), len(c)) for x, c in zip(zip(*rows), columns))
+    row_template = ("|" + (" %%-%ds |" * len(columns))) % widths
     header = row_template % tuple(columns)
-    bar = '+%s+' % '+'.join('-' * (w + 2) for w in widths)
-    data = '\n'.join(row_template % r for r in rows)
-    return '\n'.join([bar, header, bar, data, bar])
+    bar = "+%s+" % "+".join("-" * (w + 2) for w in widths)
+    data = "\n".join(row_template % r for r in rows)
+    return "\n".join([bar, header, bar, data, bar])
 
 
 def put_lines(buf, lines):
-    if any(not isinstance(x, unicode) for x in lines):
-        lines = [unicode(x) for x in lines]
-    buf.write('\n'.join(lines))
+    if any(not isinstance(x, str) for x in lines):
+        lines = [str(x) for x in lines]
+    buf.write("\n".join(lines))
 
 
 _method_cache = {}
@@ -679,7 +866,7 @@ class methodcaller(object):
     serializable
     """
 
-    __slots__ = ('method',)
+    __slots__ = ("method",)
     func = property(lambda self: self.method)  # For `funcname` to work
 
     def __new__(cls, method):
@@ -709,7 +896,8 @@ class itemgetter(object):
     Unlike the builtin `operator.itemgetter`, instances of this class are
     serializable
     """
-    __slots__ = ('index',)
+
+    __slots__ = ("index",)
 
     def __init__(self, index):
         self.index = index
@@ -734,6 +922,7 @@ class MethodCache(object):
     >>> M.count(a, 3) == a.count(3)
     True
     """
+
     __getattr__ = staticmethod(methodcaller)
     __dir__ = lambda self: list(_method_cache)
 
@@ -769,6 +958,7 @@ class SerializableLock(object):
 
     The creation of locks is itself not threadsafe.
     """
+
     def __init__(self, token=None):
         self.token = token or str(uuid.uuid4())
         if self.token in SerializableLock._locks:
@@ -777,11 +967,11 @@ class SerializableLock(object):
             self.lock = Lock()
             SerializableLock._locks[self.token] = self.lock
 
-    def acquire(self, *args):
-        return self.lock.acquire(*args)
+    def acquire(self, *args, **kwargs):
+        return self.lock.acquire(*args, **kwargs)
 
-    def release(self, *args):
-        return self.lock.release(*args)
+    def release(self, *args, **kwargs):
+        return self.lock.release(*args, **kwargs)
 
     def __enter__(self):
         self.lock.__enter__()
@@ -789,9 +979,8 @@ class SerializableLock(object):
     def __exit__(self, *args):
         self.lock.__exit__(*args)
 
-    @property
     def locked(self):
-        return self.locked
+        return self.lock.locked()
 
     def __getstate__(self):
         return self.token
@@ -810,8 +999,8 @@ def get_scheduler_lock(collection=None, scheduler=None):
        scheduler used."""
     from . import multiprocessing
     from .base import get_scheduler
-    actual_get = get_scheduler(collections=[collection],
-                               scheduler=scheduler)
+
+    actual_get = get_scheduler(collections=[collection], scheduler=scheduler)
 
     if actual_get == multiprocessing.get:
         return multiprocessing.get_context().Manager().Lock()
@@ -822,10 +1011,14 @@ def get_scheduler_lock(collection=None, scheduler=None):
 def ensure_dict(d):
     if type(d) is dict:
         return d
-    elif hasattr(d, 'dicts'):
+    elif hasattr(d, "dicts"):
         result = {}
+        seen = set()
         for dd in d.dicts.values():
-            result.update(dd)
+            dd_id = id(dd)
+            if dd_id not in seen:
+                result.update(dd)
+                seen.add(dd_id)
         return result
     return dict(d)
 
@@ -838,24 +1031,24 @@ class OperatorMethodMixin(object):
         """ bind operator to this class """
         name = op.__name__
 
-        if name.endswith('_'):
+        if name.endswith("_"):
             # for and_ and or_
             name = name[:-1]
-        elif name == 'inv':
-            name = 'invert'
+        elif name == "inv":
+            name = "invert"
 
-        meth = '__{0}__'.format(name)
+        meth = "__{0}__".format(name)
 
-        if name in ('abs', 'invert', 'neg', 'pos'):
-            bind_method(cls, meth, cls._get_unary_operator(op))
+        if name in ("abs", "invert", "neg", "pos"):
+            setattr(cls, meth, cls._get_unary_operator(op))
         else:
-            bind_method(cls, meth, cls._get_binary_operator(op))
+            setattr(cls, meth, cls._get_binary_operator(op))
 
-            if name in ('eq', 'gt', 'ge', 'lt', 'le', 'ne', 'getitem'):
+            if name in ("eq", "gt", "ge", "lt", "le", "ne", "getitem"):
                 return
 
-            rmeth = '__r{0}__'.format(name)
-            bind_method(cls, rmeth, cls._get_binary_operator(op, inv=True))
+            rmeth = "__r{0}__".format(name)
+            setattr(cls, rmeth, cls._get_binary_operator(op, inv=True))
 
     @classmethod
     def _get_unary_operator(cls, op):
@@ -875,8 +1068,8 @@ def partial_by_order(*args, **kwargs):
     >>> partial_by_order(5, function=add, other=[(1, 10)])
     15
     """
-    function = kwargs.pop('function')
-    other = kwargs.pop('other')
+    function = kwargs.pop("function")
+    other = kwargs.pop("other")
     args2 = list(args)
     for i, arg in other:
         args2.insert(i, arg)
@@ -889,8 +1082,9 @@ def is_arraylike(x):
     Examples
     --------
     >>> import numpy as np
-    >>> x = np.ones(5)
-    >>> is_arraylike(x)
+    >>> is_arraylike(np.ones(5))
+    True
+    >>> is_arraylike(np.ones(()))
     True
     >>> is_arraylike(5)
     False
@@ -899,10 +1093,40 @@ def is_arraylike(x):
     """
     from .base import is_dask_collection
 
+    return bool(
+        hasattr(x, "shape")
+        and isinstance(x.shape, tuple)
+        and hasattr(x, "dtype")
+        and not any(is_dask_collection(n) for n in x.shape)
+    )
+
+
+def is_dataframe_like(df):
+    """ Looks like a Pandas DataFrame """
+    typ = type(df)
     return (
-        hasattr(x, 'shape') and x.shape and
-        hasattr(x, 'dtype') and
-        not any(is_dask_collection(n) for n in x.shape)
+        all(hasattr(typ, name) for name in ("groupby", "head", "merge", "mean"))
+        and all(hasattr(df, name) for name in ("dtypes", "columns"))
+        and not any(hasattr(typ, name) for name in ("name", "dtype"))
+    )
+
+
+def is_series_like(s):
+    """ Looks like a Pandas Series """
+    typ = type(s)
+    return (
+        all(hasattr(typ, name) for name in ("groupby", "head", "mean"))
+        and all(hasattr(s, name) for name in ("dtype", "name"))
+        and "index" not in typ.__name__.lower()
+    )
+
+
+def is_index_like(s):
+    """ Looks like a Pandas Index """
+    typ = type(s)
+    return (
+        all(hasattr(s, name) for name in ("name", "dtype"))
+        and "index" in typ.__name__.lower()
     )
 
 
@@ -932,8 +1156,7 @@ def natural_sort_key(s):
     >>> sorted(a, key=natural_sort_key)
     ['f0', 'f1', 'f2', 'f8', 'f9', 'f10', 'f11', 'f19', 'f20', 'f21']
     """
-    return [int(part) if part.isdigit() else part
-            for part in re.split(r'(\d+)', s)]
+    return [int(part) if part.isdigit() else part for part in re.split(r"(\d+)", s)]
 
 
 def factors(n):
@@ -966,12 +1189,16 @@ def parse_bytes(s):
     1000000000
     >>> parse_bytes('MB')
     1000000
+    >>> parse_bytes(123)
+    123
     >>> parse_bytes('5 foos')  # doctest: +SKIP
     ValueError: Could not interpret 'foos' as a byte unit
     """
-    s = s.replace(' ', '')
+    if isinstance(s, (int, float)):
+        return int(s)
+    s = s.replace(" ", "")
     if not s[0].isdigit():
-        s = '1' + s
+        s = "1" + s
 
     for i in range(len(s) - 1, -1, -1):
         if not s[i].isalpha():
@@ -996,32 +1223,150 @@ def parse_bytes(s):
 
 
 byte_sizes = {
-    'kB': 10**3,
-    'MB': 10**6,
-    'GB': 10**9,
-    'TB': 10**12,
-    'PB': 10**15,
-    'KiB': 2**10,
-    'MiB': 2**20,
-    'GiB': 2**30,
-    'TiB': 2**40,
-    'PiB': 2**50,
-    'B': 1,
-    '': 1,
+    "kB": 10 ** 3,
+    "MB": 10 ** 6,
+    "GB": 10 ** 9,
+    "TB": 10 ** 12,
+    "PB": 10 ** 15,
+    "KiB": 2 ** 10,
+    "MiB": 2 ** 20,
+    "GiB": 2 ** 30,
+    "TiB": 2 ** 40,
+    "PiB": 2 ** 50,
+    "B": 1,
+    "": 1,
 }
 byte_sizes = {k.lower(): v for k, v in byte_sizes.items()}
-byte_sizes.update({k[0]: v for k, v in byte_sizes.items() if k and 'i' not in k})
-byte_sizes.update({k[:-1]: v for k, v in byte_sizes.items() if k and 'i' in k})
+byte_sizes.update({k[0]: v for k, v in byte_sizes.items() if k and "i" not in k})
+byte_sizes.update({k[:-1]: v for k, v in byte_sizes.items() if k and "i" in k})
+
+
+def format_time(n):
+    """ format integers as time
+
+    >>> format_time(1)
+    '1.00 s'
+    >>> format_time(0.001234)
+    '1.23 ms'
+    >>> format_time(0.00012345)
+    '123.45 us'
+    >>> format_time(123.456)
+    '123.46 s'
+    """
+    if n >= 1:
+        return "%.2f s" % n
+    if n >= 1e-3:
+        return "%.2f ms" % (n * 1e3)
+    return "%.2f us" % (n * 1e6)
+
+
+def format_bytes(n):
+    """ Format bytes as text
+
+    >>> format_bytes(1)
+    '1 B'
+    >>> format_bytes(1234)
+    '1.23 kB'
+    >>> format_bytes(12345678)
+    '12.35 MB'
+    >>> format_bytes(1234567890)
+    '1.23 GB'
+    >>> format_bytes(1234567890000)
+    '1.23 TB'
+    >>> format_bytes(1234567890000000)
+    '1.23 PB'
+    """
+    if n > 1e15:
+        return "%0.2f PB" % (n / 1e15)
+    if n > 1e12:
+        return "%0.2f TB" % (n / 1e12)
+    if n > 1e9:
+        return "%0.2f GB" % (n / 1e9)
+    if n > 1e6:
+        return "%0.2f MB" % (n / 1e6)
+    if n > 1e3:
+        return "%0.2f kB" % (n / 1000)
+    return "%d B" % n
+
+
+timedelta_sizes = {
+    "s": 1,
+    "ms": 1e-3,
+    "us": 1e-6,
+    "ns": 1e-9,
+    "m": 60,
+    "h": 3600,
+    "d": 3600 * 24,
+}
+
+tds2 = {
+    "second": 1,
+    "minute": 60,
+    "hour": 60 * 60,
+    "day": 60 * 60 * 24,
+    "millisecond": 1e-3,
+    "microsecond": 1e-6,
+    "nanosecond": 1e-9,
+}
+tds2.update({k + "s": v for k, v in tds2.items()})
+timedelta_sizes.update(tds2)
+timedelta_sizes.update({k.upper(): v for k, v in timedelta_sizes.items()})
+
+
+def parse_timedelta(s, default="seconds"):
+    """ Parse timedelta string to number of seconds
+
+    Examples
+    --------
+    >>> parse_timedelta('3s')
+    3
+    >>> parse_timedelta('3.5 seconds')
+    3.5
+    >>> parse_timedelta('300ms')
+    0.3
+    >>> parse_timedelta(timedelta(seconds=3))  # also supports timedeltas
+    3
+    """
+    if s is None:
+        return None
+    if isinstance(s, timedelta):
+        s = s.total_seconds()
+        return int(s) if int(s) == s else s
+    if isinstance(s, Number):
+        s = str(s)
+    s = s.replace(" ", "")
+    if not s[0].isdigit():
+        s = "1" + s
+
+    for i in range(len(s) - 1, -1, -1):
+        if not s[i].isalpha():
+            break
+    index = i + 1
+
+    prefix = s[:index]
+    suffix = s[index:] or default
+
+    n = float(prefix)
+
+    multiplier = timedelta_sizes[suffix.lower()]
+
+    result = n * multiplier
+    if int(result) == result:
+        result = int(result)
+    return result
 
 
 def has_keyword(func, keyword):
     try:
-        if PY3:
-            return keyword in inspect.signature(func).parameters
-        else:
-            if isinstance(func, functools.partial):
-                return keyword in inspect.getargspec(func.func).args
-            else:
-                return keyword in inspect.getargspec(func).args
+        return keyword in inspect.signature(func).parameters
     except Exception:
         return False
+
+
+def ndimlist(seq):
+    if not isinstance(seq, (list, tuple)):
+        return 0
+    elif not seq:
+        return 1
+    else:
+        return 1 + ndimlist(seq[0])
